"""
test.py:

A receive-ordered history that relies on predicates to provide filtering, searching, and
retrieval operations

:author: koran
"""
from fprime_gds.common.history.history import History
from fprime_gds.common.testing_fw import predicates


class TestHistory(History):
    """
    A receive-ordered history to support the GDS test api. This history adds support for specifying
    start with predicates and python's bracket notation.
    """

    __test__ = False
    ###########################################################################
    #   History Functions
    ###########################################################################

    def __init__(self, filter_pred=None):
        """
        Constructor used to set-up history. If the history is given a filter, it will ignore (drop)
        objects that don't satisfy the filter predicate.

        Args:
            filter_pred: an optional predicate to filter incoming data_objects
        """
        self.objects = []

        self.filter = predicates.always_true()
        if filter_pred is not None:
            if predicates.is_predicate(filter_pred):
                self.filter = filter_pred
            else:
                raise TypeError("The given filter was an instance of predicate.")

        self.retrieved_cursor = 0

    def data_callback(self, data, sender=None):
        """
        Data callback to push an object on the history. This callback will only add data_objects
        that satisfy the filter predicate.

        Args:
            data: object to store
        """
        if self.filter(data):
            self.objects.append(data)

    def retrieve(self, start=None):
        """
        Retrieve objects from this history. If a starting point is specified, will return a
        sub-list of all objects beginning at start to the latest object.
        Note: if no item satisfies the start predicate or the index is greater than the length of
        the history, an empty list will be returned.

        Args:
            start: optional first object to retrieve. can either be an index (int) or a predicate.
        Returns:
            a list of objects in chronological order
        """
        index = self.__get_index(start) if start is not None else 0
        self.retrieved_cursor = self.size()
        return self.objects[index:]

    def retrieve_new(self):
        """
        Retrieves a chronological order of objects that haven't been accessed through retrieve or
        retrieve_new before.

        Returns:
            a list of objects in chronological order
        """
        index = self.retrieved_cursor
        self.retrieved_cursor = self.size()
        return self.objects[index:]

    def clear(self, start=None):
        """
        Clears objects from history. A clear that specifies a starting point will clear the history
        such that start becomes the earliest element in the history after objects are removed. If
        the start is specified as a predicate, start will be the earliest object to satisfy the
        predicate.
        Note: if no item satisfies the start predicate or the index is greater than the length of
        the history, all items will be cleared.

        Args:
            start: clear all objects before start. start can either be an index or a predicate.
        """
        index = self.__get_index(start) if start is not None else self.size()

        self.retrieved_cursor -= index
        self.retrieved_cursor = max(self.retrieved_cursor, 0)

        del self.objects[:index]

    def size(self):
        """
        Accessor for the number of objects in the history
        Returns:
            the number of objects (int)
        """
        return len(self.objects)

    ###########################################################################
    #   Python Special Methods
    ###########################################################################
    def __len__(self):
        """
        Accessor for the number of objects in the history
        Returns:
            the number of objects
        """
        return self.size()

    def __getitem__(self, index):
        """
        __get_item__ is a special method in python that allows using brackets.
        Example: item = history[2] # this would return the second item in the history.

        Args:
            index: the index of the array to return.
        Returns:
            the item at the index specified.
        """
        return self.objects[index]

    def __get_index(self, start):
        """
        finds the index that start specifies
        Args:
            start: an indicator of a position in an order can be a predicate or an index in
                the ordering
        Returns:
            the index in the given list that start refers to
        """
        if not predicates.is_predicate(start):
            return start
        index = 0
        while index < self.size() and not start(self.objects[index]):
            index += 1
        return index



"""
prm_json_loader.py:

Loads flight dictionary (JSON) and returns id and mnemonic based Python dictionaries of params

@author zimri.leisher
"""

from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException


class PrmJsonLoader(JsonLoader):
    """Class to load parameters from json dictionaries"""

    PARAMS_FIELD = "parameters"

    ID = "id"
    NAME = "name"
    TYPE = "type"
    DESC = "annotation"
    DEFAULT = "default"
    

    def construct_dicts(self, _):
        """
        Constructs and returns python dictionaries keyed on id and name

        Args:
            _: Unused argument (inherited)
        Returns:
            A tuple with two channel dictionaries (python type dict):
            (id_dict, fqn_name_dict). The keys should be the channels' id and
            fully qualified name fields respectively and the values should be PrmTemplate
            objects.
        """
        id_dict = {}
        fqn_name_dict = {}

        if self.PARAMS_FIELD not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Ground Dictionary missing '{self.PARAMS_FIELD}' field: {str(self.json_file)}"
            )

        for prm_dict in self.json_dict[self.PARAMS_FIELD]:
            # Create a channel template object
            prm_temp = self.construct_template_from_dict(prm_dict)

            id_dict[prm_temp.get_id()] = prm_temp
            fqn_name_dict[prm_temp.get_full_name()] = prm_temp

        return (
            dict(sorted(id_dict.items())),
            dict(sorted(fqn_name_dict.items())),
            self.get_versions(),
        )

    def construct_template_from_dict(self, prm_dict: dict) -> PrmTemplate:
        try:
            prm_id = prm_dict[self.ID]
            # The below assignment also raises a ValueError if the name does not contain a '.'
            qualified_component_name, prm_name = prm_dict[self.NAME].rsplit('.', 1)
            if not qualified_component_name or not prm_name:
                raise ValueError()

            type_obj = self.parse_type(prm_dict[self.TYPE])
        except ValueError as e:
            raise GdsDictionaryParsingException(
                f"Parameter dictionary entry malformed, expected name of the form '<QUAL_COMP_NAME>.<PRM_NAME>' in : {str(prm_dict)}"
            )
        except KeyError as e:
            raise GdsDictionaryParsingException(
                f"{str(e)} key missing from parameter dictionary entry or its associated type in the dictionary: {str(prm_dict)}"
            )

        prm_default_val = prm_dict.get(self.DEFAULT, None)

        return PrmTemplate(
            prm_id,
            prm_name,
            qualified_component_name,
            type_obj,
            prm_default_val
        )


"""
@brief Enumeration for event severities

@date Created July 19, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from enum import Enum


class EventSeverity(Enum):
    COMMAND = 1
    ACTIVITY_LO = 2
    ACTIVITY_HI = 3
    WARNING_LO = 4
    WARNING_HI = 5
    DIAGNOSTIC = 6
    FATAL = 7



import filecmp
from pathlib import Path
import tempfile
from fprime_gds.common.tools.params import convert_json


def assert_prmdb_cfg(input: Path, expected_output: Path = None, should_fail=False):
    dict_file = Path(__file__).parent / "resources" / "simple_dictionary.json"

    temp_dir = tempfile.TemporaryDirectory()
    output_bin = Path(f"{temp_dir.name}/out_binary")
    if should_fail:
        try:
            convert_json(input, dict_file, output_bin, "dat")
            temp_dir.cleanup()
            raise RuntimeError("Config file conversion did not fail")
        except BaseException as e:
            # failed, test succeeded
            temp_dir.cleanup()
            return
    else:
        if expected_output is None:
            raise RuntimeError("Must specify expected output if should fail is False")
        convert_json(input, dict_file, output_bin, "dat")
    is_equal = filecmp.cmp(output_bin, expected_output)
    temp_dir.cleanup()

    assert is_equal


def test_nominal_paramdb():
    expected_output = Path(__file__).parent / "expected" / "simple_paramdb.dat"
    cfg_file = Path(__file__).parent / "input" / "simple_paramdb.json"
    assert_prmdb_cfg(cfg_file, expected_output)


def test_failure_paramdb():
    cfg_file = Path(__file__).parent / "input" / "simple_bad_paramdb.json"
    assert_prmdb_cfg(cfg_file, should_fail=True)



"""
@brief Packet Template class

Instances of this class describe a telemetry packet. For example an instance may
describe the packet with ID 5 and channels A, B, and C in that order.

@date Created July 2, 2018
@date Modified July 12, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize.type_exceptions import TypeMismatchException

from . import ch_template, data_template


class PktTemplate(data_template.DataTemplate):
    """Class to create packet templates to describe specific packet types"""

    def __init__(self, pkt_id, pkt_name, ch_temp_list):
        """
        Constructor

        Args:
            pkt_id (int): The ID of the telemetry packet being described
            pkt_name (str): Packet name
            ch_temp_list (ch_template list): List of ch_template objects
                         describing the channels included in the packer. The
                         order of the list is the order of the channels in the
                         packet.
        """
        super().__init__()
        if not isinstance(pkt_id, int):
            raise TypeMismatchException(int, type(pkt_id))

        if not isinstance(pkt_name, str):
            raise TypeMismatchException(str, type(pkt_name))

        if not isinstance(ch_temp_list, list):
            raise TypeMismatchException(list, type(ch_temp_list))

        for ch in ch_temp_list:
            if not isinstance(ch, ch_template.ChTemplate):
                raise TypeMismatchException(type(ch_template.ChTemplate), type(ch))

        self.id = pkt_id
        self.name = pkt_name
        self.ch_list = ch_temp_list

    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_ch_list(self):
        return self.ch_list



"""
router.py:

Sets up a router object to handle incoming and outgoing messages adding in handshake token to the front of the message
that can then be sent back as a handshake packet. This router will allow returning handshake callback end up at the
originating object correctly.

@author mstarch
"""
import struct

import fprime_gds.common.handlers


class OutgoingRouter(
    fprime_gds.common.handlers.DataHandler, fprime_gds.common.handlers.HandlerRegistrar
):
    """
    Handshake router that inspects the originating caller and notes it in a table. This table can then be consulted when
    the data returns for sending handshakes back to the originating object.
    """

    def __init__(self):
        """
        Construct the outgoing router with a dictionary to map IDs
        """
        super().__init__()

    def data_callback(self, data, sender=None):
        """
        Handles incoming data by stamping on a handshake token to be passed back in the handshake packet. This reads
        from the sender parameter to create this token, otherwise "0000" is sent out.

        :param data: encoded data to be prepended to
        :param sender: sender to append to.
        """
        if sender is None:
            sender = 0
        elif isinstance(sender, int):
            raise ValueError("Sender expected to integer")
        bytes_data = data + struct.unpack(">I", sender)
        self.send_to_all(bytes_data)


class IncomingRouter(
    fprime_gds.common.handlers.DataHandler, fprime_gds.common.handlers.HandlerRegistrar
):
    """
    Handshake router that inspects the returning handshake packet, and routes back to the originating object found in
    the outgoing router.
    """

    def data_callback(self, data, sender=None):
        """
        Handles incoming data by stamping on a handshake token to be passed back in the handshake packet. This reads
        from the sender parameter to create this token, otherwise "0000" is sent out.

        :param data: encoded data to be prepended to
        :param sender: sender id to append to.
        """
        if sender is None:
            sender = 0
        elif isinstance(sender, int):
            raise ValueError("Sender expected to integer")
        bytes_data = data + struct.unpack(">I", sender)
        self.send_to_all(bytes_data)



from __future__ import annotations
import typing
from typing import Any

# This makes the code forward-compatible. In Python 3.10+, the `|` operator
# creates a types.UnionType. In 3.9, only typing.Union exists.
try:
    from types import UnionType

    UNION_TYPES = (typing.Union, UnionType)
except ImportError:
    UNION_TYPES = (typing.Union,)

from dataclasses import dataclass, fields, astuple
from typing import ClassVar
import typing
from typing import Union
from pathlib import Path
import struct
import zlib
from fprime.common.models.serialize.type_base import BaseType
from fprime.common.models.serialize.numerical_types import (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I64Type,
    I16Type,
    I32Type,
    I8Type,
    F32Type,
    F64Type,
)
from fprime.common.models.serialize.bool_type import BoolType
from enum import Enum


def get_union_members(type_hint: type) -> list[type]:
    """
    If the type_hint is a Union, returns a list of its member types.
    Otherwise, returns the original type_hint.
    """
    # get_origin returns the base type (e.g., Union for Union[int, str])
    # or None if it's a simple type like int.
    origin = typing.get_origin(type_hint)

    if origin in UNION_TYPES:
        # get_args returns the type arguments (e.g., (int, str))
        return list(typing.get_args(type_hint))

    # Not a Union, so return the type itself
    return [type_hint]


FwSizeType = U64Type
FwChanIdType = U32Type
FwPrmIdType = U32Type
FwOpcodeType = U32Type


class DirectiveId(Enum):
    INVALID = 0
    WAIT_REL = 1
    WAIT_ABS = 2
    GOTO = 4
    IF = 5
    NO_OP = 6
    STORE_TLM_VAL = 7
    STORE_PRM = 8
    CONST_CMD = 9
    # stack op directives
    # all of these are handled at the CPP level by one StackOpDirective
    # boolean ops
    OR = 10
    AND = 11
    # integer equalities
    IEQ = 12
    INE = 13
    # unsigned integer inequalities
    ULT = 14
    ULE = 15
    UGT = 16
    UGE = 17
    # signed integer inequalities
    SLT = 18
    SLE = 19
    SGT = 20
    SGE = 21
    # floating point equalities
    FEQ = 22
    FNE = 23
    # floating point inequalities
    FLT = 24
    FLE = 25
    FGT = 26
    FGE = 27
    NOT = 28
    # floating point conversion to signed/unsigned integer,
    # and vice versa
    FPTOSI = 29
    FPTOUI = 30
    SITOFP = 31
    UITOFP = 32
    # integer arithmetic
    IADD = 33
    ISUB = 34
    IMUL = 35
    UDIV = 36
    SDIV = 37
    UMOD = 38
    SMOD = 39
    # float arithmetic
    FADD = 40
    FSUB = 41
    FMUL = 42
    FDIV = 43
    FLOAT_FLOOR_DIV = 44
    FPOW = 45
    FLOG = 46
    FMOD = 47
    # floating point bitwidth conversions
    FPEXT = 48
    FPTRUNC = 49
    # integer bitwidth conversions
    # signed integer extend
    SIEXT_8_64 = 50
    SIEXT_16_64 = 51
    SIEXT_32_64 = 52
    # zero (unsigned) integer extend
    ZIEXT_8_64 = 53
    ZIEXT_16_64 = 54
    ZIEXT_32_64 = 55
    # integer truncate
    ITRUNC_64_8 = 56
    ITRUNC_64_16 = 57
    ITRUNC_64_32 = 58
    # end stack op dirs

    EXIT = 59
    ALLOCATE = 60
    STORE = 61
    LOAD = 62
    PUSH_VAL = 63
    DISCARD = 64
    MEMCMP = 65
    STACK_CMD = 66


class Directive:
    opcode: ClassVar[DirectiveId] = DirectiveId.INVALID

    def serialize(self) -> bytes:
        arg_bytes = self.serialize_args()

        output = U8Type(self.opcode.value).serialize()
        output += U16Type(len(arg_bytes)).serialize()
        output += arg_bytes

        return output

    def serialize_args(self) -> bytes:
        output = bytes()

        for field in fields(self):
            value = getattr(self, field.name)
            if isinstance(value, BaseType):
                # it is already an fprime type instance
                # so we can serialize it
                output += value.serialize()
                continue

            if isinstance(value, bytes):
                # it is just raw bytes
                output += value
                continue

            # okay, it is not a primitive type or bytes
            field_type = typing.get_type_hints(self.__class__)[field.name]
            union_members = get_union_members(field_type)
            primitive_type = None
            # find out which primitive type it is
            for arg in union_members:
                if issubclass(arg, BaseType):
                    # it is a primitive type
                    primitive_type = arg
                    break
            if primitive_type is None:
                raise NotImplementedError(
                    "Unknown how to serialize field", field.name, "for", self
                )

            output += primitive_type(value).serialize()

        return output

    def __repr__(self):
        r = self.__class__.__old_repr__(self)
        name = self.__class__.__name__.replace("Directive", "").upper()
        value = "".join(r.split("(")[1:])
        return name + "(" + value

    @classmethod
    def deserialize(cls, data: bytes, offset: int) -> tuple[int, "Directive"] | None:
        if len(data) - offset < 3:
            # insufficient space
            return None
        opcode = struct.unpack_from(">B", data, offset)[0]
        arg_size = struct.unpack_from(">H", data, offset + 1)[0]
        offset += 3
        if len(data) - offset < arg_size:
            # insufficient space
            return None
        args = data[offset : (offset + arg_size)]
        offset += arg_size
        dir_type = [
            c
            for c in (Directive.__subclasses__() + StackOpDirective.__subclasses__())
            if c.opcode.value == opcode
        ]
        if len(dir_type) != 1:
            return None

        arg_offset = 0
        dir_type = dir_type[0]
        arg_values = []

        # go through each field in the type of the directive
        for field in fields(dir_type):
            field_type = typing.get_type_hints(dir_type)[field.name]
            # get a list of all union members of the field type
            # or a list containing just the type if it is not a union
            union_types = get_union_members(field_type)

            base_type = None
            for t in union_types:
                if issubclass(t, BaseType):
                    base_type = t

            # if one of the members of the union was a sub of basetype
            if base_type is not None:
                # deserialize using that basetype and add to arg value list
                instance = base_type()
                instance.deserialize(args, arg_offset)
                arg_values.append(instance.val)
                arg_offset += instance.getSize()
                continue
            # none of the args were base types. the only other thing we could be
            # is a byte array. assert that that's true
            assert len(union_types) == 1 and union_types[0] == bytes
            # it is just raw bytes. deserialize until the end
            arg_values.append(args[arg_offset:])
            arg_offset = len(args)
            continue

        dir = dir_type(*arg_values)
        return offset, dir

@dataclass
class StackOpDirective(Directive):
    """the argument types this dir pops off the stack"""

    stack_output_type: ClassVar[type[BaseType]] = BaseType
    """the type this dir pushes to the stack"""


@dataclass
class StackCmdDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STACK_CMD

    args_size: Union[int, U32Type]


@dataclass
class MemCompareDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.MEMCMP
    size: Union[int, U32Type]


@dataclass
class LoadDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.LOAD

    lvar_offset: Union[int, U32Type]
    size: Union[int, U32Type]


@dataclass
class IntegerSignedExtend8To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SIEXT_8_64
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class IntegerSignedExtend16To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SIEXT_16_64
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class IntegerSignedExtend32To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SIEXT_32_64
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class IntegerZeroExtend8To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ZIEXT_8_64
    stack_output_type: ClassVar[type[BaseType]] = U64Type


@dataclass
class IntegerZeroExtend16To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ZIEXT_16_64
    stack_output_type: ClassVar[type[BaseType]] = U64Type


@dataclass
class IntegerZeroExtend32To64Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ZIEXT_32_64
    stack_output_type: ClassVar[type[BaseType]] = U64Type


@dataclass
class IntegerTruncate64To8Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ITRUNC_64_8
    stack_output_type: ClassVar[type[BaseType]] = I8Type


@dataclass
class IntegerTruncate64To16Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ITRUNC_64_16
    stack_output_type: ClassVar[type[BaseType]] = I16Type


@dataclass
class IntegerTruncate64To32Directive(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ITRUNC_64_32
    stack_output_type: ClassVar[type[BaseType]] = I32Type


@dataclass
class AllocateDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.ALLOCATE

    size: Union[int, U32Type]


@dataclass
class StoreDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE

    lvar_offset: Union[int, U32Type]
    size: Union[int, U32Type]


@dataclass
class DiscardDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.DISCARD

    size: Union[int, U32Type]


@dataclass
class PushValDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.PUSH_VAL

    val: bytes


@dataclass
class ConstCmdDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.CONST_CMD

    cmd_opcode: Union[int, FwOpcodeType]
    args: bytes


@dataclass
class FloatModuloDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FMOD
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class SignedModuloDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SMOD
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class UnsignedModuloDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UMOD
    stack_output_type: ClassVar[type[BaseType]] = U64Type


@dataclass
class IntAddDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.IADD
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class IntSubtractDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ISUB
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class IntMultiplyDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.IMUL
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class UnsignedIntDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UDIV
    stack_output_type: ClassVar[type[BaseType]] = U64Type


@dataclass
class SignedIntDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SDIV
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class FloatAddDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FADD
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatSubtractDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FSUB
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatMultiplyDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FMUL
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatExponentDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPOW
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FDIV
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatFloorDivideDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLOAT_FLOOR_DIV
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatLogDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLOG
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class WaitRelDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.WAIT_REL
    # seconds and useconds are implicit


@dataclass
class WaitAbsDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.WAIT_ABS
    # time base, time context, seconds and useconds are implicit


@dataclass
class GotoDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.GOTO
    dir_idx: Union[int, U32Type]


@dataclass
class IfDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.IF
    false_goto_dir_index: Union[int, U32Type]
    """U32: The dir index to go to if the top of stack is false."""


@dataclass
class NoOpDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.NO_OP


@dataclass
class StoreTlmValDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE_TLM_VAL
    chan_id: Union[int, FwChanIdType]
    """FwChanIdType: The telemetry channel ID to get."""
    lvar_offset: Union[int, U32Type]


@dataclass
class StorePrmDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.STORE_PRM
    prm_id: Union[int, FwPrmIdType]
    """FwPrmIdType: The parameter ID to get the value of."""
    lvar_offset: Union[int, U32Type]


@dataclass
class OrDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.OR
    stack_args: ClassVar[list[type[BaseType]]] = [BoolType, BoolType]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class AndDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.AND
    stack_args: ClassVar[list[type[BaseType]]] = [BoolType, BoolType]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class IntEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.IEQ
    stack_args: ClassVar[list[type[BaseType]]] = [
        Union[I64Type, U64Type],
        Union[I64Type, U64Type],
    ]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class IntNotEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.INE
    stack_args: ClassVar[list[type[BaseType]]] = [
        Union[I64Type, U64Type],
        Union[I64Type, U64Type],
    ]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class UnsignedLessThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ULT
    stack_args: ClassVar[list[type[BaseType]]] = [U64Type, U64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class UnsignedLessThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.ULE
    stack_args: ClassVar[list[type[BaseType]]] = [U64Type, U64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class UnsignedGreaterThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UGT
    stack_args: ClassVar[list[type[BaseType]]] = [U64Type, U64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class UnsignedGreaterThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UGE
    stack_args: ClassVar[list[type[BaseType]]] = [U64Type, U64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class SignedLessThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SLT
    stack_args: ClassVar[list[type[BaseType]]] = [I64Type, I64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class SignedLessThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SLE
    stack_args: ClassVar[list[type[BaseType]]] = [I64Type, I64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class SignedGreaterThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SGT
    stack_args: ClassVar[list[type[BaseType]]] = [I64Type, I64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class SignedGreaterThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SGE
    stack_args: ClassVar[list[type[BaseType]]] = [I64Type, I64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatGreaterThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FGE
    stack_args: ClassVar[list[type[BaseType]]] = [F64Type, F64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatLessThanOrEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLE
    stack_args: ClassVar[list[type[BaseType]]] = [F64Type, F64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatLessThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FLT
    stack_args: ClassVar[list[type[BaseType]]] = [F64Type, F64Type]
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatGreaterThanDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FGT
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FEQ
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatNotEqualDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FNE
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class NotDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.NOT
    stack_output_type: ClassVar[type[BaseType]] = BoolType


@dataclass
class FloatTruncateDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPTRUNC
    stack_output_type: ClassVar[type[BaseType]] = F32Type


@dataclass
class FloatExtendDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPEXT
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatToSignedIntDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPTOSI
    stack_output_type: ClassVar[type[BaseType]] = I64Type


@dataclass
class SignedIntToFloatDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.SITOFP
    stack_output_type: ClassVar[type[BaseType]] = F64Type


@dataclass
class FloatToUnsignedIntDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.FPTOUI
    stack_output_type: ClassVar[type[BaseType]] = U64Type


@dataclass
class UnsignedIntToFloatDirective(StackOpDirective):
    opcode: ClassVar[DirectiveId] = DirectiveId.UITOFP
    stack_output_type: ClassVar[type[BaseType]] = F64Type
    # src implied


@dataclass
class ExitDirective(Directive):
    opcode: ClassVar[DirectiveId] = DirectiveId.EXIT


for cls in Directive.__subclasses__() + StackOpDirective.__subclasses__():
    cls.__old_repr__ = cls.__repr__
    cls.__repr__ = Directive.__repr__


class UnaryStackOp(str, Enum):
    NOT = "not"
    IDENTITY = "+"
    NEGATE = "-"


class BinaryStackOp(str, Enum):
    EXPONENT = "**"
    MODULUS = "%"
    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    FLOOR_DIVIDE = "//"
    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="
    LESS_THAN_OR_EQUAL = "<="
    LESS_THAN = "<"
    EQUAL = "=="
    NOT_EQUAL = "!="
    OR = "or"
    AND = "and"


NUMERIC_OPERATORS = {
    UnaryStackOp.IDENTITY,
    UnaryStackOp.NEGATE,
    BinaryStackOp.ADD,
    BinaryStackOp.SUBTRACT,
    BinaryStackOp.MULTIPLY,
    BinaryStackOp.DIVIDE,
    BinaryStackOp.MODULUS,
    BinaryStackOp.EXPONENT,
    BinaryStackOp.FLOOR_DIVIDE,
}
BOOLEAN_OPERATORS = {UnaryStackOp.NOT, BinaryStackOp.OR, BinaryStackOp.AND}

UNARY_STACK_OPS: dict[str, dict[type[BaseType], type[StackOpDirective]]] = {
    UnaryStackOp.NOT: {BoolType: NotDirective},
    UnaryStackOp.IDENTITY: {
        I64Type: NoOpDirective,
        U64Type: NoOpDirective,
        F64Type: NoOpDirective,
    },
    UnaryStackOp.NEGATE: {
        I64Type: IntMultiplyDirective,
        U64Type: IntMultiplyDirective,
        F64Type: FloatMultiplyDirective
    },
}

BINARY_STACK_OPS: dict[str, dict[type[BaseType], type[StackOpDirective]]] = {
    BinaryStackOp.EXPONENT: {F64Type: FloatExponentDirective},
    BinaryStackOp.MODULUS: {
        I64Type: SignedModuloDirective,
        U64Type: UnsignedModuloDirective,
        F64Type: FloatModuloDirective,
    },
    BinaryStackOp.ADD: {
        I64Type: IntAddDirective,
        U64Type: IntAddDirective,
        F64Type: FloatAddDirective,
    },
    BinaryStackOp.SUBTRACT: {
        I64Type: IntSubtractDirective,
        U64Type: IntSubtractDirective,
        F64Type: FloatSubtractDirective,
    },
    BinaryStackOp.MULTIPLY: {
        I64Type: IntMultiplyDirective,
        U64Type: IntMultiplyDirective,
        F64Type: FloatMultiplyDirective,
    },
    BinaryStackOp.DIVIDE: {
        I64Type: SignedIntDivideDirective,
        U64Type: UnsignedIntDivideDirective,
        F64Type: FloatDivideDirective,
    },
    BinaryStackOp.FLOOR_DIVIDE: {
        I64Type: SignedIntDivideDirective,
        U64Type: UnsignedIntDivideDirective,
        F64Type: FloatFloorDivideDirective,
    },
    BinaryStackOp.GREATER_THAN: {
        I64Type: SignedGreaterThanDirective,
        U64Type: UnsignedGreaterThanDirective,
        F64Type: FloatGreaterThanDirective,
    },
    BinaryStackOp.GREATER_THAN_OR_EQUAL: {
        I64Type: SignedGreaterThanOrEqualDirective,
        U64Type: UnsignedGreaterThanOrEqualDirective,
        F64Type: FloatGreaterThanOrEqualDirective,
    },
    BinaryStackOp.LESS_THAN_OR_EQUAL: {
        I64Type: SignedLessThanOrEqualDirective,
        U64Type: UnsignedLessThanOrEqualDirective,
        F64Type: FloatLessThanOrEqualDirective,
    },
    BinaryStackOp.LESS_THAN: {
        I64Type: SignedLessThanDirective,
        U64Type: UnsignedLessThanDirective,
        F64Type: FloatLessThanDirective,
    },
    BinaryStackOp.EQUAL: {
        I64Type: IntEqualDirective,
        U64Type: IntEqualDirective,
        F64Type: FloatEqualDirective,
    },
    BinaryStackOp.NOT_EQUAL: {
        I64Type: IntNotEqualDirective,
        U64Type: IntNotEqualDirective,
        F64Type: FloatNotEqualDirective,
    },
    BinaryStackOp.OR: {BoolType: OrDirective},
    BinaryStackOp.AND: {BoolType: AndDirective},
}



"""
histories.py:

Module used to handle the wrangling of histories for the standard pipeline. This allows the standard pipeline, and other
to compose in this code.

@author mstarch
"""

from typing import Type

from fprime_gds.common.history.history import History
from fprime_gds.common.history.ram import RamHistory


class Histories:
    """
    Class to handle the individual histories. This handles the following histories:

    1. Channel history
    2. Event history
    3. Command history (short-circuited feedback from encoder)
    """

    def __init__(self):
        """Constructor of histories composer"""
        self.coders = None
        self._command_hist = None
        self._event_hist = None
        self._channel_hist = None
        self._implementation_type = RamHistory

    def setup_histories(self, coders):
        """
        Setup a set of history objects in order to store the events of the decoders. This registers itself with the
        supplied coders object.

        :param coders: coders object to register histories with
        """
        self.coders = coders
        # Allow implementation type to disable histories
        if self._implementation_type is None:
            return
        # Create histories, RAM histories for now
        self.commands = self._implementation_type()
        self.events = self._implementation_type()
        self.channels = self._implementation_type()

    @property
    def implementation(self):
        """Get implementation type"""
        return self._implementation_type

    @implementation.setter
    def implementation(self, implementation_type: Type[History]):
        """Set the implementation type"""
        assert (
            self._command_hist is None
            and self._event_hist is None
            and self._channel_hist is None
        ), "Cannot setup implementation types after setup"
        self._implementation_type = implementation_type

    @property
    def events(self):
        """
        Events history property
        """
        return self._event_hist

    @events.setter
    def events(self, history: History):
        """
        Set the events history
        """
        assert (
            self.coders is not None
        ), "Cannot override history before calling 'setup_histories'"
        if self._event_hist is None:
            self.coders.remove_event_consumer(self._event_hist)
        self._event_hist = history
        self.coders.register_event_consumer(self._event_hist)

    @property
    def channels(self):
        """
        Channels history property
        """
        return self._channel_hist

    @channels.setter
    def channels(self, history: History):
        """
        Set the channels history
        """
        assert (
            self.coders is not None
        ), "Cannot override history before calling 'setup_histories'"
        if self._channel_hist is None:
            self.coders.remove_channel_consumer(self._channel_hist)
        self._channel_hist = history
        self.coders.register_channel_consumer(self._channel_hist)

    @property
    def commands(self):
        """
        Commands history property
        """
        return self._command_hist

    @commands.setter
    def commands(self, history: History):
        """
        Set the channels history
        """
        assert (
            self.coders is not None
        ), "Cannot override history before calling 'setup_histories'"
        if self._command_hist is None:
            self.coders.remove_command_consumer(self._command_hist)
        self._command_hist = history
        self.coders.register_command_consumer(self._command_hist)



####
#
####
import re
import sys
from io import StringIO
from pathlib import Path

import flask_restful
import flask_restful.reqparse

from fprime_gds.common.tools.seqgen import SeqGenException, generateSequence


class StdioTheif(object):
    """
    This class consumes all standard out and error production produced with-in a context block (with :) capturing it.
    """

    def __init__(self):
        """Setup our capture devices"""
        self.io = StringIO()
        self.output = ""

    def write(self, *args, **kwargs):
        """Write tee for all possible inputs"""
        stripped_msg = " ".join(args)
        self.io.write(stripped_msg, **kwargs)

    def __enter__(self):
        """Entry overwrite"""
        sys.stdout = self
        sys.stderr = self
        return None

    def __exit__(self, *_):
        """No worries here"""
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        self.io.seek(0)
        self.output = self.io.read()


class SequenceCompiler(flask_restful.Resource):
    def __init__(self, dictionary, tempdir, uplinker, destination):
        self.dictionary = dictionary
        self.tempdir = Path(tempdir)
        self.uplinker = uplinker
        self.destination = destination

        self.parser = flask_restful.reqparse.RequestParser()
        self.parser.add_argument(
            "key", required=True, help="Protection key. Must be: 0xfeedcafe."
        )
        self.parser.add_argument(
            "name", required=True, help="Name of sequence file to create"
        )
        self.parser.add_argument(
            "text", required=True, help="Text of sequence file to create"
        )
        self.parser.add_argument(
            "uplink", required=True, help="Text of sequence file to create"
        )

    def put(self):
        args = self.parser.parse_args()
        key = args.get("key", None)
        name = args.get("name", "")
        text = args.get("text", "")
        uplink = args.get("uplink", "false") == "true"
        if key is None or int(key, 0) != 0xFEEDCAFE:
            flask_restful.abort(
                403,
                message=f"{key} is invalid command key. Supply 0xfeedcafe to run command.",
            )
        elif not re.match(".*\\.seq", name) or Path(name).name != name:
            flask_restful.abort(
                403,
                message={"error": "Supply filename with .seq suffix", "type": "error"},
            )
        temp_seq_path = self.tempdir / Path(name).name
        temp_bin_path = temp_seq_path.with_suffix(".bin")
        messages = ""
        try:
            with open(temp_seq_path, "w") as file_handle:
                file_handle.write(text)
            thief = StdioTheif()
            with thief:
                generateSequence(
                    temp_seq_path, temp_bin_path, self.dictionary, 0xFFFF, cont=True
                )
            messages += thief.output
            if uplink:
                destination = Path(self.destination) / temp_bin_path.name
                self.uplinker.enqueue(str(temp_bin_path), str(destination))
                messages += f"Uplinking to {destination}. Please confirm uplink EVRs before running.\n"
        except OSError as ose:
            flask_restful.abort(403, message={"error": str(ose), "type": "error"})
        except SeqGenException as exc:
            flask_restful.abort(403, message={"error": str(exc), "type": "validation"})
        if temp_seq_path.exists():
            temp_seq_path.unlink()
        return {"message": messages}



from __future__ import annotations
from dataclasses import dataclass, field, fields
from numbers import Number
from pathlib import Path
from typing import Union
from lark import Lark, Token, Transformer, v_args
from lark.tree import Meta

from fprime_gds.common.fpy.bytecode.directives import Directive, StackOpDirective
from fprime.common.models.serialize.type_base import BaseType
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.numerical_types import NumericalType

fpybc_grammar_str = (Path(__file__).parent / "grammar.lark").read_text()


def parse(text: str):
    parser = Lark(
        fpybc_grammar_str,
        start="input",
        parser="lalr",
        propagate_positions=True,
        maybe_placeholders=True,
    )

    tree = parser.parse(text)
    transformed = FpyBcTransformer().transform(tree)
    return transformed


@dataclass
class Node:
    meta: Meta = field(repr=False)
    id: int = field(init=False, repr=False, default=None)

    def __hash__(self):
        return hash(self.id)


@dataclass
class NodeOpWithNoArgs(Node):
    op: str


@dataclass
class NodeOpWithArgs(Node):
    op: str
    args: list[Node]


@dataclass
class NodeGotoTag(Node):
    tag: str


@dataclass
class NodeBytes(Node):
    value: int | str


NodeStmt = Union[NodeGotoTag, NodeOpWithNoArgs, NodeOpWithArgs]


@dataclass
class NodeBody(Node):
    stmts: list[NodeStmt]


@v_args(meta=False, inline=False)
def as_list(self, tree):
    return list(tree)


def no_inline_or_meta(type):
    @v_args(meta=False, inline=False)
    def wrapper(self, tree):
        return type(tree)

    return wrapper


def no_inline(type):
    @v_args(meta=True, inline=False)
    def wrapper(self, meta, tree):
        return type(meta, tree)

    return wrapper


def no_meta(type):
    @v_args(meta=False, inline=True)
    def wrapper(self, tree):
        return type(tree)

    return wrapper


def handle_str(meta, s: str):
    return s.strip("'").strip('"')

def handle_bytes(meta, value: list[int|str]|None):
    if value is None:
        return bytes()

    ret = bytes()
    for val in value:
        if isinstance(val, int):
            if val > 255 or val < 0:
                raise RuntimeError("Each byte must be < 256 and >= 0")
            ret += val.to_bytes(1, "little")
        elif isinstance(val, str):
            ret += val.encode("utf-8")

    return ret


def handle_op_with_args(meta, tree: list[Token]):
    return NodeOpWithArgs(meta, tree[0].value, [t for t in tree[1:] if t is not None])


def handle_op_with_no_args(meta, tree: list[Token]):
    return NodeOpWithNoArgs(meta, tree[0].value)


def handle_hex(self, s: str):
    return int(s, 16)


@v_args(meta=True, inline=True)
class FpyBcTransformer(Transformer):
    input = no_inline(NodeBody)

    goto_tag = NodeGotoTag
    op_with_no_args = no_inline(handle_op_with_no_args)
    op_with_args = no_inline(handle_op_with_args)

    NAME = str
    DEC_NUMBER = int
    HEX_NUMBER = handle_hex
    bytes = no_inline(handle_bytes)
    FLOAT_NUMBER = float
    STRING = no_inline(handle_str)
    CONST_TRUE = lambda a, b: True
    CONST_FALSE = lambda a, b: False
    OP_WITH_NO_ARGS = str


def assemble(body: NodeBody) -> tuple[bytes, int]:
    line_idx = 0
    # collect all goto tags
    tags: dict[str, int] = {}
    for stmt in body.stmts:
        if isinstance(stmt, NodeGotoTag):
            tags[stmt.tag] = line_idx
        else:
            line_idx += 1

    dirs: list[Directive] = []
    for stmt in body.stmts:
        if isinstance(stmt, NodeGotoTag):
            continue

        # okay, it's a directive. which dir?

        op = stmt.op.upper()
        dir_type = None

        for subcls in Directive.__subclasses__() + StackOpDirective.__subclasses__():
            if subcls.opcode.name == op:
                dir_type = subcls
                break
        assert dir_type is not None, op

        # okay. if it's a no arg dir, just instantiate and add to list
        if isinstance(stmt, NodeOpWithNoArgs):
            dirs.append(dir_type())
        else:
            args = []
            if op == "GOTO" or op == "IF":
                if isinstance(stmt.args[0], str):
                    # lookup tag
                    tag_idx = tags.get(stmt.args[0], None)
                    if tag_idx is None:
                        raise RuntimeError(f"Unknown tag {stmt.args[0]}")
                    args.append(tag_idx)
                else:
                    args = stmt.args
            else:
                args = stmt.args

            dirs.append(dir_type(*args))

    return dirs


def directives_to_fpybc(dirs: list[Directive]) -> str:
    out = ""
    for dir in dirs:
        # write the op name
        out += dir.opcode.name.lower()

        # write the args
        for field in fields(dir):
            field_value = getattr(dir, field.name)
            val = None

            if isinstance(field_value, BaseType):
                assert isinstance(field_value, (NumericalType, StringType, BoolType)), field_value
                val = field_value.val
            else:
                val = field_value

            if isinstance(val, str):
                out += ' "' + val + '"'
            elif isinstance(val, (Number, bool)):
                out += " " + str(val)
            elif isinstance(val, bytes):
                for byte in val:
                    out += " " + str(byte)
            else:
                assert False, type(val)

        out += "\n"

    return out



"""
publishing.py:

This file contains a basic publishing pipeline. It reads and writes data bound for the GDS.

:author: lestarch
"""


from typing import Type

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.encoders.ch_encoder import ChEncoder
from fprime_gds.common.encoders.event_encoder import EventEncoder
from fprime_gds.common.transport import RoutingTag, ThreadedTCPSocketClient
from fprime_gds.common.utils.config_manager import ConfigManager

from fprime_gds.common.handlers import DataHandler, MappedRegistrar
from ..models import dictionaries


class PublishingPipeline(DataHandler, MappedRegistrar):
    """ Pipeline for publishing

    This pipeline sets up the following process:

    Data -> Encoder -> Client -> <wire>
    """
    DEFAULT_ENCODERS = {
        "FW_PACKET_LOG": EventEncoder,
        "FW_PACKET_TELEM": ChEncoder
    }

    def __init__(self):
        """
        Set core variables to None or their composition handlers.
        """
        super().__init__()
        self.client_socket = None

        self._dictionaries = dictionaries.Dictionaries()
        self._transport_type = ThreadedTCPSocketClient

    def setup(self, dictionaries):
        """ Set up the publishing pipeline """
        self._dictionaries = dictionaries
        self.client_socket = self.__transport_type()
        config = ConfigManager.get_instance()
    
        for id, encoder_class in self.DEFAULT_ENCODERS.items():
            encoder_instance = encoder_class(config=config)
            encoder_instance.register(self.client_socket)
            self.register(id, encoder_instance)

    def data_callback(self, data, sender=None):
        """ Publish data """
        if isinstance(data, ChData):
            self.send_to_all("FW_PACKET_TELEM", data)
        elif isinstance(data, EventData):
            self.send_to_all("FW_PACKET_LOG", data)
        return super().data_callback(data, sender)
    
    def publish_channel(self, name, value, time):
        """ Publish channel value using name, time, and value

        Looks up the channel template in the dictionary and constructs a new channel object given the time and value.
        This ChData object is then sent into the outgoing "publish" pipeline.

        """
        template = self._dictionaries.channel_name[name]
        copied_value_object = template.get_type_obj()(value)
        copied_value_object.val = value
        object = ChData(copied_value_object, time, template)
        return self.data_callback(object, self)

    @property
    def transport_implementation(self):
        """Get implementation type for transport"""
        return self.__transport_type

    @transport_implementation.setter
    def transport_implementation(self, transport_type: Type[None]):
        """Set the implementation type for transport"""
        assert (
            self.client_socket is None
        ), "Cannot setup transport implementation type after setup"
        self.__transport_type = transport_type


    def connect(
        self, connection_uri, incoming_tag=RoutingTag.GUI, outgoing_tag=RoutingTag.GUI
    ):
        """Connects to the middleware layer

        Connect to the middleware layer. This connection needs to identify if the object is a a FSW or GUI client. The
        default connection acts as a GUI client sending back to the GUI.

        Args:
            connection_uri: URI of the connection to make
            incoming_tag: this pipeline will act as supplied tag (GUI, FSW). Default: GUI
            outgoing_tag: this pipeline will produce data for supplied tag (FSW, GUI). Default: FSW
        """
        self.client_socket.connect(connection_uri, incoming_tag, outgoing_tag)

    def disconnect(self):
        """ Disconnect from the client socket """
        if self.client_socket is not None:
            self.client_socket.disconnect()

    



from fprime.common.models.serialize.numerical_types import U32Type
import pytest
from fprime_gds.common.fpy.test_helpers import (
    assert_run_success,
    assert_compile_failure,
    assert_compile_success,
    assert_run_failure,
    lookup_type,
)


# define this function if you want to just use the Python fpy model
@pytest.fixture(name="fprime_test_api", scope="module")
def fprime_test_api_override():
    """A file-specific override that simply returns None."""
    return None


def test_simple_var(fprime_test_api):
    seq = """
var: U32 = 1
"""

    assert_run_success(fprime_test_api, seq)


def test_int_literal(fprime_test_api):
    seq = """
var: I64 = 123_456
var = -123_456
var = +123_456
var = 000_00000_0
"""

    assert_run_success(fprime_test_api, seq)


def test_bad_int_literal(fprime_test_api):
    seq = """
var: I64 = 0123_456

"""

    assert_compile_failure(fprime_test_api, seq)

def test_float_literal(fprime_test_api):
    seq = """
var: F32 = 1.000e-5
var = .1
var = 1.
var = 2.123
var = 100.5e+10
var = -123.456
"""

    assert_run_success(fprime_test_api, seq)


def test_exit_success(fprime_test_api):
    seq = """
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_exit_failure(fprime_test_api):
    seq = """
exit(False)
"""
    assert_run_failure(fprime_test_api, seq)


def test_large_var(fprime_test_api):
    seq = """
var: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
"""

    assert_run_success(fprime_test_api, seq)


def test_var_assign_to_var(fprime_test_api):
    seq = """
x: U32 = 1
var: U32 = x
"""

    assert_run_success(fprime_test_api, seq)


def test_nonexistent_var(fprime_test_api):
    seq = """
var = 1
"""

    assert_compile_failure(fprime_test_api, seq)


def test_create_after_assign_var(fprime_test_api):
    seq = """
var = 1
var: U32 = 2
"""

    assert_compile_failure(fprime_test_api, seq)


def test_bad_assign_type(fprime_test_api):
    seq = """
var: failure = 1
"""

    assert_compile_failure(fprime_test_api, seq)


def test_weird_assign_type(fprime_test_api):
    seq = """
var: CdhCore.cmdDisp.CMD_NO_OP = 1
"""

    assert_compile_failure(fprime_test_api, seq)


def test_reassign(fprime_test_api):
    seq = """
var: U32 = 1
var = 2
"""

    assert_run_success(fprime_test_api, seq)


def test_reassign_ann(fprime_test_api):
    seq = """
var: U32 = 1
var: U32 = 2
"""
    assert_compile_failure(fprime_test_api, seq)


def test_assign_inconsistent_type(fprime_test_api):
    seq = """
var: U32 = 1
var: U16 = 2
"""

    assert_compile_failure(fprime_test_api, seq)


def test_call_cmd(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP()
"""
    assert_run_success(fprime_test_api, seq)


def test_call_cmd_with_str_arg(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP_STRING("hello world")
"""
    assert_run_success(fprime_test_api, seq)


def test_call_cmd_with_int_arg(fprime_test_api):
    seq = """
Ref.sendBuffComp.PARAMETER3_PRM_SET(4)
"""
    assert_run_success(fprime_test_api, seq)


def test_bad_enum_ctor(fprime_test_api):
    seq = """
Ref.SG5.Settings(123, 0.5, 0.5, Ref.SignalType(1))
"""
    assert_compile_failure(fprime_test_api, seq)


def test_cmd_with_enum(fprime_test_api):
    seq = """
Ref.SG5.Settings(123, 0.5, 0.5, Ref.SignalType.TRIANGLE)
"""
    assert_run_success(fprime_test_api, seq)


def test_instantiate_type_for_cmd(fprime_test_api):
    seq = """
Ref.typeDemo.CHOICE_PAIR(Ref.ChoicePair(Ref.Choice.ONE, Ref.Choice.TWO))
"""
    assert_run_success(fprime_test_api, seq)


def test_var_with_enum_type(fprime_test_api):
    seq = """
var: Ref.Choice = Ref.Choice.ONE
"""

    assert_run_success(fprime_test_api, seq)


def test_simple_if(fprime_test_api):
    seq = """
var: bool = True

# use exit(True) if we want the sequence to succeed
# exit(False) if we want it to fail. helpful for testing.

if var:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_expr(fprime_test_api):
    seq = """
if True or False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_not_expr(fprime_test_api):
    seq = """
if not False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_expr_with_vars(fprime_test_api):
    seq = """
var1: bool = True
var2: bool = False

if var1 or var2:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_geq(fprime_test_api):
    seq = """
if 2 >= 1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_geq_tlm(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP()
if CdhCore.cmdDisp.CommandsDispatched >= 1:
    exit(True)
exit(False)
"""

    assert_run_success(
        fprime_test_api,
        seq,
        {"CdhCore.cmdDisp.CommandsDispatched": U32Type(1).serialize()},
    )


def test_large_elifs(fprime_test_api):
    seq = """
if CdhCore.cmdDisp.CommandsDispatched == 0:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("0")
elif CdhCore.cmdDisp.CommandsDispatched == 1:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("1")
elif CdhCore.cmdDisp.CommandsDispatched == 2:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("2")
elif CdhCore.cmdDisp.CommandsDispatched == 3:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("3")
elif CdhCore.cmdDisp.CommandsDispatched == 4:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("4")
else:
    CdhCore.cmdDisp.CMD_NO_OP_STRING(">4")
"""

    assert_run_success(
        fprime_test_api,
        seq,
        {"CdhCore.cmdDisp.CommandsDispatched": U32Type(4).serialize()},
    )


def test_int_as_stmt(fprime_test_api):
    seq = """
2
"""

    assert_compile_failure(fprime_test_api, seq)


def test_complex_as_stmt(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP
"""

    assert_compile_failure(fprime_test_api, seq)


def test_get_struct_member(fprime_test_api):
    seq = """
if Ref.cmdSeq.Debug.nextStatementOpcode == 0:
    # should be 0 because we aren't in debug mode
    exit(True)
exit(False)
"""

    assert_run_success(
        fprime_test_api,
        seq,
        {
            "Ref.cmdSeq.Debug": lookup_type(
                fprime_test_api, "Svc.FpySequencer.DebugTelemetry"
            )(
                {
                    "reachedEndOfFile": False,
                    "nextStatementReadSuccess": False,
                    "nextStatementOpcode": 0,
                    "nextCmdOpcode": 0,
                }
            ).serialize()
        },
    )


def test_get_const_struct_member(fprime_test_api):
    seq = """
var: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
if var.priority == 3:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_get_const_member_of_ctor(fprime_test_api):
    seq = """
# currently this is not supported, but it should be in the future
var: U32 = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED).priority
if var == 3:
    exit(True)
exit(False)
"""

    assert_compile_failure(fprime_test_api, seq)


def test_float_cmp(fprime_test_api):
    seq = """
if 4.0 > 5.0:
    exit(False)
exit(True)
"""

    assert_run_success(fprime_test_api, seq)


def test_wait_rel(fprime_test_api):
    seq = """
sleep(1, 1000)
"""
    assert_run_success(fprime_test_api, seq)


def test_f32_f64_cmp(fprime_test_api):
    seq = """
val: F32 = 0.0
val2: F64 = 1.0
if val > val2:
    exit(False)
exit(True)
"""

    assert_run_success(fprime_test_api, seq)


def test_construct_array(fprime_test_api):
    seq = """
val: Svc.ComQueueDepth = Svc.ComQueueDepth(0, 0)
"""

    assert_run_success(fprime_test_api, seq)


def test_get_item_of_var(fprime_test_api):
    seq = """
val: Svc.ComQueueDepth = Svc.ComQueueDepth(0, 0)
if val[0] == 0:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_i32_f64_cmp(fprime_test_api):
    seq = """
val: I32 = 2
val2: F64 = 1.0
if val > val2:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_i32_u32_cmp(fprime_test_api):
    seq = """
val: I32 = -2
val2: U32 = 2
# this is actually false because we interpret both sides as unsigned
if val < val2:
    exit(False)
exit(True)
"""

    assert_run_success(fprime_test_api, seq)


# caught one bug
def test_float_int_literal_cmp(fprime_test_api):
    seq = """
if 1 < 2.0:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_assign_float_to_int(fprime_test_api):
    seq = """
val: I64 = 1.0
"""

    assert_compile_failure(fprime_test_api, seq)


# caught one bug
def test_and_of_ors(fprime_test_api):
    seq = """
if True or False and True or True:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_if_true(fprime_test_api):
    seq = """
if True:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_false(fprime_test_api):
    seq = """
if False:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_else_true(fprime_test_api):
    seq = """
if True:
    exit(True)
else:
    exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_else_false(fprime_test_api):
    seq = """
if False:
    exit(False)
else:
    exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_elif_else(fprime_test_api):
    seq = """
if False:
    exit(False)
elif True:
    exit(True)
else:
    exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_and_true_true(fprime_test_api):
    seq = """
if True and True:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_and_true_false(fprime_test_api):
    seq = """
if True and False:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_false_false(fprime_test_api):
    seq = """
if False or False:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_true_false(fprime_test_api):
    seq = """
if True or False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_not_true(fprime_test_api):
    seq = """
if not True:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_not_false(fprime_test_api):
    seq = """
if not False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_complex_and_or_not(fprime_test_api):
    seq = """
if not False and (True or False):
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_literal_comparison(fprime_test_api):
    seq = """
if 255 > 254:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_literal_comparison_false(fprime_test_api):
    seq = """
if 255 < 254:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_u8(fprime_test_api):
    seq = """
val1: U8 = 200
val2: U8 = 100

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_i8(fprime_test_api):
    seq = """
val1: I8 = 100
val2: I8 = -100

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_u32(fprime_test_api):
    seq = """
val1: U32 = 4294967295
val2: U32 = 0

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_i32(fprime_test_api):
    seq = """
val1: I32 = 2147483647
val2: I32 = -2147483648

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_f32(fprime_test_api):
    seq = """
val1: F32 = 3.14159
val2: F32 = -3.14159

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_f64(fprime_test_api):
    seq = """
val1: F64 = 3.14159265359
val2: F64 = -3.14159265359

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_mixed_numeric_comparisons(fprime_test_api):
    seq = """
val_u8: U8 = 255
val_i8: I8 = -10
val_u32: U32 = 4294967295
val_i32: I32 = -2147483648
val_f32: F32 = 3.14159
val_f64: F64 = -3.14159265359

# i32 > u32 because the cmp happens as unsigned, and so the
# two's complement negative is really large
if val_u8 < val_i8 and val_i32 > val_u32:
    if val_f64 <= val_f32 and val_f32 >= val_f64:
        if val_u8 != val_i8 and not (val_u32 == val_i32):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_equality_edge_cases(fprime_test_api):
    seq = """
val1: U8 = 0
val2: U8 = 0
val3: F32 = 0.0
val4: F64 = 0.0
val5: I32 = 0

if val1 == val2 and val3 == val4 and val4 == val5:
    if not (val1 != val2) and not (val3 != val4) and not (val4 != val5):
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_nested_boolean_expressions(fprime_test_api):
    seq = """
if not (True and False or True and not False) and True:
    exit(False)  # Should not execute
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_maximum_integer_comparisons(fprime_test_api):
    seq = """
val1: U64 = 18446744073709551615  # Max U64
val2: I64 = 9223372036854775807   # Max I64, should be same in unsigned
# TODO there is currently a bug in this
#val3: I64 = -9223372036854775808  # Min I64, should be max i64 + 1 in unsigned
val3: I64 = -9223372036854775807  # Min I64 - 1, should be max i64 + 1 in unsigned

if val1 > val2 and val2 > val3:
    if val3 < val1:
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_complex_type_assignments(fprime_test_api):
    seq = """
val1: I8 = 127
val2: U8 = 255
val3: F32 = 127.0

if val1 == val3:  # Integer to float comparison
    if val2 > val3:  # Unsigned vs float comparison
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_negative_val_unsigned_type(fprime_test_api):
    seq = """
val1: U32 = -1  # Should succeed and be equal to largest u32 val
exit(val1 == 2 ** 32 - 1)
"""
    assert_run_success(fprime_test_api, seq)


def test_overflow_compile_error(fprime_test_api):
    seq = """
val1: U8 = 256  # Should fail: value too large for U8
"""
    assert_compile_failure(fprime_test_api, seq)


def test_mixed_boolean_numeric_comparison(fprime_test_api):
    seq = """
val1: U8 = 1
val2: I8 = -1
if (val1 > 0) == True and (val2 < 0) == True:  # Compare boolean results
    if not ((val1 <= 0) == True or (val2 >= 0) == True):
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_complex_boolean_nesting(fprime_test_api):
    seq = """
if not not not not not True:  # Multiple not operators
    exit(False)
elif not (True and not (False or not True)):  # Complex nesting
    exit(False)
else:
    exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_non_const_str_arg(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP_STRING(Ref.cmdSeq.SeqPath)
"""
    # currently can't do non const string args
    assert_compile_failure(fprime_test_api, seq)


def test_non_const_int_arg(fprime_test_api):
    seq = """
var: U8 = 255
Ref.sendBuffComp.PARAMETER3_PRM_SET(var)
"""
    assert_run_success(fprime_test_api, seq)


def test_non_const_float_arg(fprime_test_api):
    seq = """
var: F32 = 1.2
Ref.sendBuffComp.PARAMETER4_PRM_SET(var)
"""
    assert_run_success(fprime_test_api, seq)


def test_non_const_builtin_arg(fprime_test_api):
    seq = """
var: U32 = 1
var2: U32 = 123123
sleep(var, var2)
"""
    assert_run_success(fprime_test_api, seq)


def test_add_unsigned(fprime_test_api):
    seq = """
var1: U32 = 500
var2: U32 = 1000
if var1 + var2 == 1500 and (var1 + 1) > var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_add_signed(fprime_test_api):
    seq = """
var1: I32 = -255
var2: I32 = 255
if var1 + var2 == 0 and (var1 + 1) > (var1 + -1):
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_add_float(fprime_test_api):
    seq = """
var1: F32 = -255.0
var2: F32 = 255.0
if var1 + var2 == 0 and (var1 + 1) > (var1 + -1):
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


# this test inspired by a bug
def test_float_truncate_stack_size(fprime_test_api):
    seq = """
var2: F64 = 123.0
var1: F32 = -var2
exit(var1 == -123.0)
"""
    assert_run_success(fprime_test_api, seq)


def test_sub_unsigned(fprime_test_api):
    seq = """
var1: U32 = 1000
var2: U32 = 500
if var1 - var2 == 500 and (var1 - 1) < var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_sub_signed(fprime_test_api):
    seq = """
var1: I32 = 255
var2: I32 = 255
if var1 - var2 == 0 and (var1 - 1) < (var1 - -1):
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_sub_float(fprime_test_api):
    seq = """
var1: F32 = 255.0
var2: F32 = 255.0
if var1 - var2 == 0 and (var1 - 1) < (var1 - -1):
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_mul_unsigned(fprime_test_api):
    seq = """
var1: U32 = 5
var2: U32 = 20
if var1 * var2 == 100 and (var1 * 2) > var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_mul_signed(fprime_test_api):
    seq = """
var1: I32 = -5
var2: I32 = 20
if var1 * var2 == -100 and (var1 * 2) < var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_mul_float(fprime_test_api):
    seq = """
var1: F32 = 5.0
var2: F32 = 20.0
if var1 * var2 == 100 and (var1 * 2) > var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_div_unsigned(fprime_test_api):
    seq = """
var1: U32 = 20
var2: U32 = 5
if var1 / var2 == 4 and (var1 / 2) < var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_div_signed(fprime_test_api):
    seq = """
var1: I32 = -20
var2: I32 = 5
if var1 / var2 == -4: # and (var1 / -2) > var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_div_float(fprime_test_api):
    seq = """
var1: F32 = -20.0
var2: F32 = 5.0
if var1 / var2 == -4 and (var1 / -2) > var1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


# this test caught one bug (my mom spotted it)
def test_order_of_operations(fprime_test_api):
    seq = """
if 1 - 2 + 3 * 4 == 11 and 10 / 5 * 2 == 4:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_arithmetic_arg_to_builtin_bad_type(fprime_test_api):
    seq = """
sleep(1 + 2 * 0, (0 + 1 / 2))
"""
    assert_compile_failure(fprime_test_api, seq)


def test_arithmetic_arg_to_builtin(fprime_test_api):
    seq = """
sleep(1 + 2 * 0, (0 + 1 // 2))
"""
    assert_run_success(fprime_test_api, seq)


def test_chain_mul(fprime_test_api):
    seq = """
var1: I32 = 1
var2: I32 = 2
var3: I32 = 3
if var1 * var2 * var3 == 6:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_chain_add(fprime_test_api):
    seq = """
var1: I32 = 1
var2: I32 = 2
var3: I32 = 3
if var1 + var2 + var3 == 6:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_chain_sub(fprime_test_api):
    seq = """
var1: I32 = 1
var2: I32 = 2
var3: I32 = 3
if var1 - var2 - var3 == -4:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_chain_div(fprime_test_api):
    seq = """
var1: I32 = 3
var2: I32 = 2
var3: I32 = 1
if var1 / var3 / var2 == 3/2:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_pow_unsigned(fprime_test_api):
    seq = """
var1: U32 = 20
var2: U32 = 2
if var1 ** var2 == 400:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_pow_signed(fprime_test_api):
    seq = """
var1: I32 = -20
var2: I32 = 2
if var1 ** var2 == 400:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_pow_float(fprime_test_api):
    seq = """
var1: F32 = 4.0
var2: F32 = 0.5
if var1 ** var2 == 2:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_int_literal_as_float(fprime_test_api):
    seq = """
var: F32 = 1
exit(var == 1.0)
"""

    assert_run_success(fprime_test_api, seq)


def test_log(fprime_test_api):
    seq = """
if log(4.0) > 1.385 and log(4.0) < 1.387:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_assign_complex(fprime_test_api):
    seq = """
var: I64 = 1 + 1
var = var + 3
if var == 5:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_assign_cycle(fprime_test_api):
    seq = """
var: I64 = var
"""
    assert_compile_failure(fprime_test_api, seq)


def test_assign_cycle_2(fprime_test_api):
    seq = """
var: I64 = (var + 1)
"""
    assert_compile_failure(fprime_test_api, seq)


def test_use_before_declare(fprime_test_api):
    seq = """
var: I64 = var2
var2: I64 = 0
"""
    assert_compile_failure(fprime_test_api, seq)


def test_math_after_cmd(fprime_test_api):
    seq = """
var: I32 = 1
CdhCore.cmdDisp.CMD_NO_OP()
# making sure that the cmd doesn't mess with the stack
if var + 1 == 2:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_cmd_return_val(fprime_test_api):
    seq = """
ret: Fw.CmdResponse = CdhCore.cmdDisp.CMD_NO_OP()
if ret == Fw.CmdResponse.OK:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_struct_eq(fprime_test_api):
    seq = """
var: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
var2: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
var3: Svc.DpRecord = Svc.DpRecord(123, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
exit(var == var2 and var != var3)
"""

    assert_run_success(fprime_test_api, seq)


def test_complex_eq_fail(fprime_test_api):
    seq = """
var: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
var2: Fw.CmdResponse = Fw.CmdResponse.OK
exit(var == var2)
"""

    assert_compile_failure(fprime_test_api, seq)


def test_mod_float(fprime_test_api):
    seq = """
var1: F32 = 25.25
var2: F32 = 5
exit(var1 % var2 == 0.25 and (var1 + 1) % var2 == 1.25)
"""
    assert_run_success(fprime_test_api, seq)


def test_mod_unsigned(fprime_test_api):
    seq = """
var1: U32 = 5
var2: U32 = 20
exit(var2 % var1 == 0 and (var2 + 1) % var1 == 1)
"""
    assert_run_success(fprime_test_api, seq)


def test_mod_signed(fprime_test_api):
    seq = """
var1: I32 = -5
var2: I32 = 20
exit(var2 % var1 == 0 and (var2 + 1) % var1 == -4)
"""
    assert_run_success(fprime_test_api, seq)


def test_bool_stack_value(fprime_test_api):
    seq = """
exit((1 == 1) == True)
"""
    assert_run_success(fprime_test_api, seq)


def test_u8_too_large(fprime_test_api):
    seq = """
var: U8 = 123
var = 256
"""
    assert_compile_failure(fprime_test_api, seq)


def test_string_eq(fprime_test_api):
    seq = """
exit("asdf" == "asdf")
"""
    assert_compile_failure(fprime_test_api, seq)


def test_string_var_eq(fprime_test_api):
    seq = """
var: string = "test"
var1: string = "test"
exit(var == var1)
"""
    assert_compile_failure(fprime_test_api, seq)


def test_string_type(fprime_test_api):
    seq = """
var: string = "test"
"""
    assert_compile_failure(fprime_test_api, seq)


def test_too_many_dirs(fprime_test_api):
    from fprime_gds.common.fpy.codegen import MAX_DIRECTIVES_COUNT

    seq = "CdhCore.cmdDisp.CMD_NO_OP()\n" * (MAX_DIRECTIVES_COUNT + 1)
    assert_compile_failure(fprime_test_api, seq)


def test_dir_too_large(fprime_test_api):
    # TODO this doesn't actually crash cuz the dir is too large... not sure at the moment how to trigger this
    from fprime_gds.common.fpy.codegen import MAX_DIRECTIVE_SIZE

    seq = 'CdhCore.cmdDisp.CMD_NO_OP_STRING("' + "a" * MAX_DIRECTIVE_SIZE + '")'
    assert_compile_failure(fprime_test_api, seq)


def test_readme_examples(fprime_test_api):
    seq = """
Ref.recvBuffComp.PARAMETER4_PRM_SET(1 - 2 + 3 * 4 + 10 / 5 * 2)
param4: F32 = 15.0
Ref.recvBuffComp.PARAMETER4_PRM_SET(param4)

prm_3: U8 = Ref.sendBuffComp.parameter3
cmds_dispatched: U32 = CdhCore.cmdDisp.CommandsDispatched

signal_pair: Ref.SignalPair = Ref.SG1.PairOutput

signal_pair_time: F32 = Ref.SG1.PairOutput.time
com_queue_depth_0: U32 = ComCcsds.comQueue.comQueueDepth[0]
value: bool = 1 > 2 and (3 + 4) != 5
many_cmds_dispatched: bool = cmds_dispatched >= 123
record1: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
record2: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
records_equal: bool = record1 == record2 # == True
random_value: I8 = 4 # chosen by fair dice roll. guaranteed to be random

if random_value < 0:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("won't happen")
elif random_value > 0 and random_value <= 6:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("should happen!")
else:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("uh oh...")
"""
    assert_compile_failure(fprime_test_api, seq)


def test_unary_plus_unsigned(fprime_test_api):
    seq = """
var: U32 = 1
exit(+var == var)
"""
    assert_run_success(fprime_test_api, seq)


def test_unary_plus_signed(fprime_test_api):
    seq = """
var: I32 = 1
exit(+var == var)
"""
    assert_run_success(fprime_test_api, seq)


def test_unary_plus_float(fprime_test_api):
    seq = """
var: F32 = 1.0
exit(+var == var)
"""
    assert_run_success(fprime_test_api, seq)


def test_unary_minus_signed(fprime_test_api):
    seq = """
var: I32 = 1
exit(-var == -1)
"""
    assert_run_success(fprime_test_api, seq)


def test_unary_minus_float(fprime_test_api):
    seq = """
var: F32 = 1.0
exit(-var == -1.0)
"""
    assert_run_success(fprime_test_api, seq)


# this is an interesting case, because one side is unsigned
# has an unsigned intermediate, so the literal is converted to
# U64, but -1 is outside range of U64 so it fails to compile.
# what should really happen here? TODO should -var fail to compile?
# should intermediate type of unary minus be signed?
def test_negative_int_literal_unsigned_op(fprime_test_api):
    seq = """
var: U32 = 1
exit(-var == -1)
"""
    assert_run_success(fprime_test_api, seq)


def test_multi_arg_variable_arg_cmd(fprime_test_api):
    seq = """
var1: I32 = 1
var2: F32 = 1.0
var3: U8 = 8
CdhCore.cmdDisp.CMD_TEST_CMD_1(var1, var2, var3)
"""
    assert_run_success(fprime_test_api, seq)



"""
handlers.py:

Defines a set of base classes that allow for the system to handle various functions of the system. Primarily this
defines the "DataHandler" base class for handling data.

@author mstarch
"""

import abc
from typing import List, Type
from fprime_gds.plugin.definitions import gds_plugin_specification


class DataHandler(abc.ABC):
    """
    Defines the necessary functions required to handle data as part of the F prime project. This allows any implementer
    to be used to handle data.
    """

    @abc.abstractmethod
    def data_callback(self, data, sender=None):
        """
        Callback function used to handle data being produced elsewhere in the system and processed by the given object.
        Data supplied should be of a known type for the given object, and sender is an id of the sender. If not supplied
        sender will be None.

        :param data: data to be handled by this class
        :param sender: (optional) id of sender, otherwise None
        """


class DataHandlerPlugin(DataHandler, abc.ABC):
    """PLugin class allowing for custom data handlers

    This class acts as a DataHandler class with the addition that it can be used as a plugin and thus self reports the
    data types it handles (whereas DataHandler leaves that up to the registration call). Users shall concretely subclass
    this class with their own data handling functionality.
    """
    def __init__(self, **kwargs):
        """ Initialize """
        self.publisher = None

    def set_publisher(self, publisher):
        """ Set publishing pipeline """
        self.publisher = publisher

    @abc.abstractmethod
    def get_handled_descriptors() -> List[str]:
        """Return a list of data descriptor names this plugin handles"""
        raise NotImplementedError()

    @classmethod
    @gds_plugin_specification
    def register_data_handler_plugin(cls) -> Type["DataHandlerPlugin"]:
        """Register a plugin to provide post-decoding data handling capabilities

        Plugin hook for registering a plugin that supplies a DataHandler implementation. Implementors of this hook must
        return a non-abstract subclass of DataHandlerPlugin. This class will be provided as a data handling
        that is automatically enabled. Users may disable this via the command line. This data handler will be supplied
        all data types returned by the `get_data_types()` method.

        This DataHandler will run within the standard GDS (UI) process. Users wanting a separate process shall use a
        GdsApp plugin instead.

        Note: users should return the class, not an instance of the class. Needed arguments for instantiation are
        determined from class methods, solicited via the command line, and provided at construction time to the chosen
        instantiation.

        Returns:
            DataHandlerPlugin subclass (not instance)
        """
        raise NotImplementedError()


class HandlerRegistrar(abc.ABC):
    """
    Defines a class that will take in registrants and remember them for calling back later. These objects should be of
    the type "DataHandler" as this handler will send data back to these handlers when asked to do so.
    """

    def __init__(self):
        """
        Constructor defining the internal lists needed to store the registrants.
        """
        super().__init__()
        self._registrants = []

    def register(self, registrant):
        """
        Register a registrant with this registrar. Will be stored and called back when asked to send data to all the
        handlers registered.

        :param registrant: handler to register
        """
        if not isinstance(registrant, DataHandler):
            raise ValueError("Cannot register non data handler")
        self._registrants.append(registrant)

    def deregister(self, registrant):
        """
        Remove a registrant from the registrar such that it will not be called back later. Note: ignores invalid
        removals by trapping the error, as the desired effect is already satisfied.

        :param registrant: registrant to remove
        :return: True if found, False if not. May safely be ignored.
        """
        try:
            self._registrants.remove(registrant)
            return True
        except ValueError:
            return False

    def send_to_all(self, data, sender=None):
        """
        Sends the given data to all registrants.

        :param data: data to send back to registrants
        :param sender: (optional) sender to pass to data_callback
        """
        for registrant in self._registrants:
            registrant.data_callback(data, sender)


class MappedRegistrar(abc.ABC):
    """ Class to register a mapping """

    def __init__(self):
        """ Initialize an empty registrant list  """
        super().__init__()
        self._registrants = {}

    def register(self, id, registrant):
        """
        Register a registrant with this registrar associated with the ID. Will be stored and called back when asked to
        send data to all the handlers registered at the id

        :param id: id to register to
        :param registrant: handler to register
        """
        self._registrants[id] = self._registrants.get(id, HandlerRegistrar())
        self._registrants[id].register(registrant)

    def deregister(self, id, registrant):
        """
        Remove a registrant from the registrar such that it will not be called back later. Note: ignores invalid
        removals by trapping the error, as the desired effect is already satisfied.

        :param id: id to register to
        :param registrant: registrant to remove
        :return: True if found, False if not. May safely be ignored.
        """
        try:
            self._registrants[id].deregister(registrant)
            return True
        except (ValueError, KeyError):
            return False

    def send_to_all(self, id, data, sender=None):
        """
        Sends the given data to all registrants at id.

        :param id: id to send to 
        :param data: data to send back to registrants
        :param sender: (optional) sender to pass to data_callback
        """
        try:
            self._registrants[id].send_to_all(data, sender)
        except KeyError:
            print("KeyERROR")
            pass


"""
encoding.py:

This file sets up the encoding and decoding handlers for use with the standard pipeline. This encapsulates the encoding
and decoding into a single component that the be composed into the standard pipeline.

@mstarch
"""

import fprime_gds.common.decoders.ch_decoder
import fprime_gds.common.decoders.event_decoder
import fprime_gds.common.decoders.file_decoder
import fprime_gds.common.decoders.pkt_decoder
import fprime_gds.common.encoders.cmd_encoder

# Encoders and Decoders
import fprime_gds.common.encoders.file_encoder


class EncodingDecoding:
    """
    Sets up and runs the encoding and decoding for the standard pipeline. This include the following encoders and
    decoders for standard setups:

    1. Events decoding
    2. Channel decoding
    3. File decoding
    4. Command encoding
    5. File encoding
    """

    def __init__(self):
        """
        Setup the encoders and decoder member variables.
        """
        # Encoder and decoder items
        self.file_encoder = None
        self.command_encoder = None
        self.event_decoder = None
        self.channel_decoder = None
        self.file_decoder = None
        self.packet_decoder = None
        self.command_subscribers = []

    def setup_coders(self, dictionaries, distributor, sender, config):
        """
        Sets up the encoder and decoder layer of the GDS pipeline. This requires a dictionary set that has loaded the
        dictionaries needed for the decoders to work correctly. This will register then register the decoders with a
        supplied distributor to handle known types with the known decoders. Lastly, the sender will be registered to the
        encoder to handle the encoded data going out.

        :param dictionaries: a dictionaries handling object holding dictionaries
        :param distributor: distributor of data to register to
        :param sender: is used to send the bytes created by the command encoder
        :param config: config object used to describe types used for different field encodings
        """
        # Create encoders and decoders using dictionaries
        self.file_encoder = fprime_gds.common.encoders.file_encoder.FileEncoder()
        self.command_encoder = fprime_gds.common.encoders.cmd_encoder.CmdEncoder(
            config=config
        )
        self.event_decoder = fprime_gds.common.decoders.event_decoder.EventDecoder(
            dictionaries.event_id, config=config
        )
        self.channel_decoder = fprime_gds.common.decoders.ch_decoder.ChDecoder(
            dictionaries.channel_id, config=config
        )
        self.file_decoder = fprime_gds.common.decoders.file_decoder.FileDecoder()
        self.packet_decoder = None
        if dictionaries.packet is not None:
            self.packet_decoder = fprime_gds.common.decoders.pkt_decoder.PktDecoder(
                dictionaries.packet, dictionaries.channel_id
            )
        # Register client socket to encoder
        self.command_encoder.register(sender)
        self.file_encoder.register(sender)
        # Register the event and channel decoders to the distributor for their
        # respective data types
        distributor.register("FW_PACKET_LOG", self.event_decoder)
        distributor.register("FW_PACKET_TELEM", self.channel_decoder)
        distributor.register("FW_PACKET_FILE", self.file_decoder)
        if self.packet_decoder is not None:
            distributor.register("FW_PACKET_PACKETIZED_TLM", self.packet_decoder)

    def send_command(self, command):
        """
        Sends a command to the registered command encoder, and further down the stream. Note: this contains a local
        loopback to any command consumers to ensure that histories and logging are updated.

        :param command: command object to send
        """
        for loopback in self.command_subscribers:
            loopback.data_callback(command)
        self.command_encoder.data_callback(command)

    def register_event_consumer(self, consumer):
        """
        Registers a history with the event decoder.

        :param consumer: consumer of events
        """
        self.event_decoder.register(consumer)

    def remove_event_consumer(self, consumer):
        """
        Removes a history from the event decoder. Will raise an error if the history was not
        previously registered.

        :param consumer: consumer of events
        :return: a boolean indicating if the consumer was removed.
        """
        return self.event_decoder.deregister(consumer)

    def register_channel_consumer(self, consumer):
        """
        Registers a history with the telemetry decoder.

        :param consumer: consumer of channels
        """
        self.channel_decoder.register(consumer)
        if self.packet_decoder is not None:
            self.packet_decoder.register(consumer)

    def remove_channel_consumer(self, consumer):
        """
        Removes a history from the telemetry decoder. Will raise an error if the history was not
        previously registered.

        :param consumer: consumer of channels
        :return: a boolean indicating if the consumer was removed.
        """
        return self.channel_decoder.deregister(consumer)

    def register_file_consumer(self, consumer):
        """
        Registers a consumer with the file decoder.

        :param consumer: consumer of file packets
        """
        self.file_decoder.register(consumer)

    def remove_file_consumer(self, consumer):
        """
        Removes a consumer from the file decoder. Will raise an error if the history was not
        previously registered.

        :param consumer: consumer of channels
        :return: a boolean indicating if the consumer was removed.
        """
        return self.file_decoder.deregister(consumer)

    def register_command_consumer(self, consumer):
        """
        Registers a history with the standard pipeline.

        :param consumer: consumer of commands
        """
        self.command_subscribers.append(consumer)

    def remove_command_consumer(self, consumer):
        """
        Removes a history that is subscribed to command data. Will raise an error if the history
        was not previously registered.

        :param consumer: consumer of commands
        :return: a boolean indicating if the consumer was removed.
        """
        try:
            self.command_subscribers.remove(consumer)
            return True
        except ValueError:
            return False

    def register_packet_consumer(self, consumer):
        """
        Registers a history with the standard pipeline.

        :param consumer: consumer of packets
        """
        if self.packet_decoder is not None:
            self.packet_decoder.register(consumer)

    def deregister_packet_consumer(self, consumer):
        """
        Removes a history that is subscribed to command data. Will raise an error if the history
        was not previously registered.

        :param consumer: consumer of packets
        :return: a boolean indicating if the consumer was removed.
        """
        if self.packet_decoder is not None:
            return self.packet_decoder.deregister(consumer)



"""
@brief Loader class for importing xml based packet dictionaries

@date Created July 12, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.data_types import exceptions

# Custom python modules
from fprime_gds.common.loaders.xml_loader import XmlLoader
from fprime_gds.common.templates.pkt_template import PktTemplate


class PktXmlLoader(XmlLoader):
    """Class to load xml packet dictionaries"""

    # Constants for use when parsing the xml
    PKT_LIST_TAG = "packets"  # tag on the root list of packets
    PKT_TAG = "packet"  # tag on each packet object
    CH_TAG = "channel"  # tag on each channel element of the packet
    NAME_FIELD = "name"  # key to retrieve the packet name
    ID_FIELD = "id"  # key to retrieve the packet id
    CH_NAME_FIELD = "name"  # key to retrieve the name from each ch elem

    def get_id_dict(self, path, ch_name_dict):
        """
        Returns the python dictionary keyed by ids for the given path

        This function will return the same dictionary originally computed for
        the given path or will construct new dictionaries if the path has never
        been passed to the get_id_dict or the get_name_dict functions.

        Args:
            path (string): Path to the xml pkt spec to convert to a python dict
            ch_name_dict (dict()): Channel dictionary with names as keys and
                                   ChTemplate objects as values.

        Returns:
            The id dictionary associated with the given path
        """
        if path in self.saved_dicts:
            (id_dict, name_dict) = self.saved_dicts[path]
        else:
            (id_dict, name_dict) = self.construct_dicts(path, ch_name_dict)
            self.saved_dicts[path] = (id_dict, name_dict)

        return id_dict

    def get_name_dict(self, path, ch_name_dict):
        """
        Returns the python dictionary keyed by names for the given path

        This function will return the same dictionary originally computed for
        the given path or will construct new dictionaries if the path has never
        been passed to the get_id_dict or the get_name_dict functions.

        Args:
            path (string): Path to the xml pkt spec to convert to a python dict
            ch_name_dict (dict()): Channel dictionary with names as keys and
                                   ChTemplate objects as values.

        Returns:
            The name dictionary associated with the given path
        """
        if path in self.saved_dicts:
            (id_dict, name_dict) = self.saved_dicts[path]
        else:
            (id_dict, name_dict) = self.construct_dicts(path, ch_name_dict)
            self.saved_dicts[path] = (id_dict, name_dict)

        return name_dict

    def construct_dicts(self, path, ch_name_dict):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            path (string): Path to the packet definition xml file to parse into
                           a dictionary.
            ch_name_dict (dict()): Channel dictionary with names as keys and
                                   ChTemplate objects as values.

        Returns:
            A tuple with two packet dictionaries (type==dict()):
            (id_dict, name_dict). The keys should be the packets' id and name
            fields respectively and the values should be PktTemplate objects.
        """
        packet_list = self.get_xml_tree(path)
        if packet_list.tag != self.PKT_LIST_TAG:
            msg = f"expected packet list to have tag {self.PKT_LIST_TAG}, but found {packet_list.tag}"
            raise exceptions.GseControllerParsingException(
                msg
            )

        id_dict = {}
        name_dict = {}

        for packet in packet_list:
            # check if this is actually a packet, and not something to ignore
            if packet.tag != self.PKT_TAG:
                continue

            pkt_name = packet.attrib[self.NAME_FIELD]
            pkt_id = int(packet.attrib[self.ID_FIELD])

            ch_list = []
            for ch in packet:
                ch_name = ch.attrib[self.CH_NAME_FIELD]

                if ch_name not in ch_name_dict:
                    msg = f"Channel {ch_name} in pkt {pkt_name}, but cannot be found in channel dictionary"
                    raise exceptions.GseControllerParsingException(
                        msg
                    )

                ch_list.append(ch_name_dict[ch_name])

            pkt_temp = PktTemplate(pkt_id, pkt_name, ch_list)

            id_dict[pkt_id] = pkt_temp
            name_dict[pkt_name] = pkt_temp

        return id_dict, name_dict



"""
ground.py:

Sets up the ground-side handlers for the comm layer. This allows the communications layer to send data and receive data
to and from the rest of the ground system. This layer consists of an Abstract base class, which guarantees the methods
available to the comm layer, and currently a single implementation used to attach to the ThreadedTcpServer.

@author lestarch
"""

import abc
import logging

from fprime_gds.common.communication.adapters.ip import TcpHandler

from .framing import TcpServerFramerDeframer

LOGGER = logging.getLogger("gds_sender")


class GroundHandler(abc.ABC):
    """
    Ground handler class interacts upstream from the comm adapter layer to the greater ground system. This
    effectively means handling the following functions:

    1. receive_all: receives any and all frames from the ground layer for uplink to the spacecraft
    2. send_all: sends any and all frames to the ground system from the spacecraft's downlink
    """

    @abc.abstractmethod
    def open(self):
        """
        Opens any needed resources and prepares the system for receiving and sending.
        """

    @abc.abstractmethod
    def receive_all(self):
        """
        Receive all packet available from the ground layer. This will return full ground packets up to the uplinker.
        These packets should be fully-deframed and ready for reframing in the comm-layer specified format.

        :return: list deframed packets
        """

    @abc.abstractmethod
    def send_all(self, frames):
        """
        Receive all packet available from the ground layer. This will return full ground packets up to the uplinker.
        These packets should be fully-deframed and ready for reframing in the comm-layer specified format.

        :return: list deframed packets
        """


class TCPGround(GroundHandler):
    """
    Interface class defining necessary functions to talk to the GDS.
    """

    def __init__(self, address="127.0.0.1", port=50050):
        """
        Initialize this interface with the address and port needed to connect to the GDS.

        :param address: Address of the tcp server. Default 127.0.0.1
        :param port: port of the tcp server. Default: 50000
        """
        self.tcp = TcpHandler(
            address, port, False, LOGGER, post_connect=b"Register FSW\n"
        )
        self.data = bytearray()
        self.deframer = TcpServerFramerDeframer()

    def open(self):
        """
        Opens any needed resources and prepares the system for receiving and sending. This means opening the TCP handler
        and sending out the initial register command to the TcpServer.
        """
        return bool(self.tcp.open())

    def close(self):
        """
        Closes the open adapter.
        """
        self.tcp.close()

    def receive_all(self):
        """
        Receive all packet available from the ground layer. This will return full ground packets up to the uplinker.
        These packets should be fully-deframed and ready for reframing in the comm-layer specified format.

        :return: list deframed packets
        """
        self.data += self.tcp.read()
        (frames, self.data, _) = self.deframer.deframe_all(self.data, no_copy=True)
        return frames

    def send_all(self, frames):
        """
        Send all packets out to the tcp socket server. This adds the framing data for the TCP Server.

        :param frames: bytes object of data to write out to the socket server
        """
        for packet in frames:
            framed = self.deframer.frame(packet)
            self.tcp.write(framed)



# ------------------------------------------------------------------------------------------
# Program: Data Product Writer
#
# Filename: data_product_writer.py
#
# Author: Garth Watney
#
#   The "Data Product Writer" program is designed to interpret and convert binary data products 
#   from the F' flight software into JSON format. Each data product comprises multiple records, 
#   and the program's role is to deserialize this binary data using JSON dictionaries and 
#   header specifications, producing a readable JSON file.
#
# Argument Parsing:
#   The program starts by parsing command-line arguments to obtain the path to the binary 
#   data product file.
#
# Reading JSON Dictionary:
#   It reads the F' JSON dictionary file, which provides the necessary context for 
#   interpreting the binary data, catching any JSON decoding errors to handle malformed 
#   files gracefully.
#
# Data Validation:
#   Before proceeding, the program checks for duplicate record identifiers within the 
#   JSON dictionary to ensure data integrity.
#
#
# Binary File Processing:
#   The program opens the binary file for reading, initializes counters for tracking the 
#   total bytes read and a variable for calculating the CRC checksum. The header data is 
#   read first, followed by the individual records, each deserialized based on the JSON 
#   dictionary and header specification.
#
# CRC Validation:
#   After processing the records, the program validates the CRC checksum to ensure the 
#   data's integrity has been maintained throughout the reading process.
#
# Exception Handling:
#   Throughout its execution, the program is equipped to handle various exceptions, 
#   ranging from file not found scenarios to specific data-related errors like CRC 
#   mismatches or duplicate record IDs. Each exception triggers an error handling 
#   routine that provides informative feedback and terminates the program gracefully.
#
# Output Generation:
#   Upon successful processing, the program writes the deserialized data to a JSON file. 
#   The output file's name is derived from the binary file's name with a change in its 
#   extension to `.json`, facilitating easy association between the input and output files.
#
# The "Data Product Writer" program is executed from the command line and requires specific 
# arguments to function correctly. The primary argument it needs is the path to the binary 
# data product file that will be processed.
#
# Usage:
# python data_product_writer.py <binFile> <json dictionary>
#
# Where:
# <binFile> is the path to the binary file generated by the F' flight software that contains 
# the data product to be deserialized and written to a JSON file.
#
# <json dictionary> is the path to the json dictionary that is generated upon an F' build
# 
#
# The program does not require any additional command-line options or flags for its basic 
# operation. Once executed with the correct binary file path and with the necessary JSON 
# files in place, it will perform the series of steps outlined previously, culminating in 
# the generation of a JSON file with the same base name as the binary file but with a .json 
# extension.
#
# ------------------------------------------------------------------------------------------

import struct
import json
import os
import sys
from typing import List, Dict, Union, ForwardRef
from pydantic import BaseModel, field_validator, computed_field, model_validator
from typing import List, Union
import argparse
from binascii import crc32

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


header_hash_data = {
    "headerHash": {
        "type": {
            "name": "U32",
            "kind": "integer",
            "signed": False,
            "size": 32
        }
    },
    "dataHash": {
        "type": {
            "name": "U32",
            "kind": "integer",
            "signed": False,
            "size": 32
        }
    }
}

# Deserialize the binary file big endian
BIG_ENDIAN = ">"


# -------------------------------------------------------------------------------------------------------------------------
# Function parse_args
#
# Description: 
#   Parse the input arguments either as passed into this function or on the command line
# -------------------------------------------------------------------------------------------------------------------------
def parse_args(args=None):
    parser = argparse.ArgumentParser(description='Data Product Writer.')
    parser.add_argument('binFile', help='Data Product Binary file')
    parser.add_argument('jsonDict', help='JSON Dictionary')
    if args is None:
        args = sys.argv[1:]
    return parser.parse_args(args)


# -------------------------------------------------------------------------------------
# These are common Pydantic classes that 
# are used by both the dictionary json and the data product json 
#
# -------------------------------------------------------------------------------------

class IntegerType(BaseModel):
    name: str
    kind: str
    size: int
    signed: bool

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "integer":
            raise ValueError('Check the "kind" field')
        return v

class FloatType(BaseModel):
    name: str
    kind: str
    size: int

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "float":
            raise ValueError('Check the "kind" field')
        return v

class BoolType(BaseModel):
    name: str
    kind: str
    size: int

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "bool":
            raise ValueError('Check the "kind" field')
        return v
    
class StringType(BaseModel):
    name: str
    kind: str
    size: int

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "string":
            raise ValueError('Check the "kind" field')
        return v

Type = ForwardRef('Type')
ArrayType = ForwardRef('ArrayType')
AliasType = ForwardRef('AliasType')

class QualifiedType(BaseModel):
    kind: str
    name: str

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "qualifiedIdentifier":
            raise ValueError('Check the "kind" field')
        return v

class StructMember(BaseModel):
    type: Union[IntegerType, FloatType, BoolType, StringType, QualifiedType]
    size: int = 1
    index: int

class StructType(BaseModel):
    kind: str
    qualifiedName: str
    members: Dict[str, StructMember]

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "struct":
            raise ValueError('Check the "kind" field')
        return v
    
class AliasType(BaseModel):
    kind: str
    qualifiedName: str
    type: Union[AliasType, StructType, ArrayType, IntegerType, BoolType, FloatType, StringType, QualifiedType]
    underlyingType: Union[StructType, ArrayType, IntegerType, BoolType, FloatType, StringType, QualifiedType]

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "alias":
            raise ValueError('Check the "kind" field')
        return v

class ArrayType(BaseModel):
    kind: str
    qualifiedName: str
    size: int
    elementType: Union[AliasType, StructType, ArrayType, IntegerType, BoolType, FloatType, StringType, QualifiedType]

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "array":
            raise ValueError('Check the "kind" field')
        return v

class Constant(BaseModel):
    qualifiedName: str
    type: Union[IntegerType, FloatType, BoolType, StringType]
    value: Union[int, float, bool, str]


class EnumeratedConstant(BaseModel):
    name: str
    value: int


class EnumType(BaseModel):
    kind: str
    qualifiedName: str
    #representationType: RepresentationType
    representationType: IntegerType
    enumeratedConstants: List[EnumeratedConstant]

    @field_validator('kind')
    def kind_qualifiedIdentifier(cls, v):
        if v != "enum":
            raise ValueError('Check the "kind" field')
        return v


class Type(BaseModel):
    type: Union[AliasType, StructType, ArrayType, IntegerType, FloatType, BoolType, QualifiedType, StringType]

class RecordStruct(BaseModel):
    name: str
    type: Union[AliasType, StructType, ArrayType, IntegerType, FloatType, BoolType, QualifiedType, StringType]
    array: bool
    id: int
    annotation: str

class ContainerStruct(BaseModel):
    name: str
    id: int
    defaultPriority: int
    annotation: str

# -------------------------------------------------------------------------------------
# These Pydantic classes define the FPRIME_DICTIONARY_FILE
#
# -------------------------------------------------------------------------------------

class FprimeDict(BaseModel):
    metadata: Dict[str, Union[str, List[str]]]
    typeDefinitions: List[Union[AliasType, ArrayType, StructType, EnumType]]
    constants: List[Constant]
    records: List[RecordStruct]
    containers: List[ContainerStruct]

# -------------------------------------------------------------------------------------
# This Pydantic class defines the data product json
#
# -------------------------------------------------------------------------------------

class DPHeader(BaseModel):
    typeDefinitions: List[Union[AliasType, ArrayType, StructType, EnumType]]
    constants: List[Constant]
    headerHash: Type
    dataHash: Type

    @computed_field
    @property
    def header(self) -> Dict[str, Union[Type, ArrayType, EnumType]]:
        # Mapping from type/constant names to header field names
        header_field_names = {
            "FwPacketDescriptorType": "PacketDescriptor",
            "FwDpIdType": "Id",
            "FwDpPriorityType" : "Priority",
            "Seconds": "Seconds",
            "USeconds": "USeconds",
            "FwTimeBaseStoreType": "TimeBase",
            "FwTimeContextStoreType": "Context",
            "Fw.DpCfg.ProcType": "ProcTypes",
            "Fw.DpCfg.CONTAINER_USER_DATA_SIZE": "UserData",
            "Fw.DpState": "DpState",
            "FwSizeStoreType": "DataSize"
        }
        # All types/constants that make up the DP header
        # Key: Values of above header_field_names
        # Values: Initialized to None, populated below based on type names/constants in the JSON dictionary
        header_dict = {v: None for v in header_field_names.values()}
        for k, v in header_field_names.items():
            # Seconds and USeconds are not in the dictionary, but are both always U32
            if k == "Seconds" or k == "USeconds":
                header_dict[v] = Type(type=IntegerType(name="U32", kind="integer", size=32, signed=False))
            # According to Fw.Dp SDD, Header::UserData is an array of U8 of size Fw::DpCfg::CONTAINER_USER_DATA_SIZE.
            elif k == "Fw.DpCfg.CONTAINER_USER_DATA_SIZE":
                for t in self.constants:
                    if t.qualifiedName == k:
                        header_dict[v] = ArrayType(
                            kind="array", 
                            qualifiedName="UserData",
                            size=t.value,
                            elementType=IntegerType(name="U8", kind="integer", size=8, signed=False)
                        )
                        break
            else:
                for t in self.typeDefinitions:
                    if t.qualifiedName == k:
                        header_dict[v] = t
                        break

        return header_dict
    
    @computed_field
    @property
    def dataId(self) -> AliasType:
        return self.header.get("Id")
    
    @computed_field
    @property
    def dataSize(self) -> AliasType:
        return self.header.get("DataSize")

    def get_size_store_bytes(self) -> int:
        return self.header.get("DataSize").underlyingType.size // 8

    @model_validator(mode='after')
    def validate_header(self) -> 'DPHeader':
        for k, v in self.header.items():
            if not v:
                raise ValueError(f'Dictionary is missing type definition or constant {k}.')
        return self

ArrayType.model_rebuild()
AliasType.model_rebuild()
StructType.model_rebuild()
Type.model_rebuild()

TypeKind = Union[AliasType, ArrayType, StructType, IntegerType, FloatType, EnumType, BoolType, QualifiedType, StringType]
TypeDef = Union[AliasType, ArrayType, StructType]

# Map the JSON types to struct format strings
type_mapping = {
    'U8': 'B',  # Unsigned 8-bit integer
    'I8': 'b',  # Signed 8-bit integer
    'U16': 'H', # Unsigned 16-bit integer
    'I16': 'h', # Signed 16 bit integer
    'U32': 'I', # Unsigned 32-bit integer
    'I32': 'i', # Signed 32-bit integer
    'I64': 'q', # Signed 64-bit integer
    'U64': 'Q', # Unsigned 64-bit integer
    'F32': 'f',  # 32-bit float
    'F64': 'd',  # 64-bit float
    'bool': '?', # An 8 bit boolean
    'string': 's'
    # Add more mappings as needed
}

# --------------------------------------------------------------------------------------------------
# class RecordIDNotFound
# 
# Description: 
#   Custom exception class for signaling the absence of a specified record ID in a JSON dictionary. 
#   It stores the missing record ID and provides a descriptive error message when invoked.
#
# Attributes:
#   recordId: The missing record ID.
#
# Methods:
#   __init__(self, recordId): Initializes the exception with the missing record ID.
#   __str__(self): Returns an error message indicating the missing record ID.
# ---------------------------------------------------------------------------------------------------

class RecordIDNotFound(Exception):
    
    def __init__(self, recordId):
        self.recordId = recordId
        
    def __str__(self):
        return f"Record ID {self.recordId} was not found in the JSON dictionary"
    
# ----------------------------------------------------------------------------------------------------------------
# class DictionaryError
# 
# Description: 
#   Custom exception class for signaling errors while parsing a JSON dictionary file.
#   It stores the file name and the line number where the error occurred, providing context in the error message.
#
# Attributes:
#   jsonFile: The name of the JSON file being parsed.
#   lineNo: The line number in the file where the error was detected.
#
# Methods:
#   __init__(self, jsonFile, lineNo): Initializes the exception with the JSON file name and the line number.
#   __str__(self): Returns an error message indicating the file and line number where the error occurred.
# ----------------------------------------------------------------------------------------------------------------
    
class DictionaryError(Exception):
    
    def __init__(self, jsonFile, lineNo):
        self.jsonFile = jsonFile
        self.lineNo = lineNo
        
    def __str__(self):
        return f"DictionaryError parsing {self.jsonFile}, line number: {self.lineNo}"

# -------------------------------------------------------------------------------------------------------------------
# class CRCError
# 
# Description: 
#   Custom exception class for signaling CRC hash mismatches during data validation.
#   It specifies whether the error occurred in the header or data, the expected CRC value, and the calculated CRC value.
#
# Attributes:
#   headerOrData: Indicates whether the mismatch occurred in the header or the data section.
#   expected: The expected CRC value.
#   calculated: The calculated CRC value that did not match the expected value.
#
# Methods:
#   __init__(self, headerOrData, expected, calculated): Initializes the exception with details of the mismatch.
#   __str__(self): Returns an error message detailing the mismatch with expected and calculated CRC values.
# --------------------------------------------------------------------------------------------------------------------
    
class CRCError(Exception):
    
    def __init__(self, headerOrData, expected, calculated):
        self.headerOrData = headerOrData
        self.expected = expected
        self.calculated = calculated
        
    def __str__(self):
        return f"CRC Hash mismatch for {self.headerOrData}: Expected {self.expected:#x}, Calculated {self.calculated:#x}"

# --------------------------------------------------------------------------------------------------------------------
# class DuplicateRecordID
#
# Description: 
#   Custom exception class for indicating the presence of a duplicate record identifier in a JSON dictionary.
#   It stores the duplicate identifier and provides a descriptive error message when invoked.
#
# Attributes:
#   identifier: The duplicate record identifier that triggered the exception.
#
# Methods:
#   __init__(self, identifier): Initializes the exception with the duplicate identifier.
#   __str__(self): Returns an error message indicating the duplicate record identifier.
# ----------------------------------------------------------------------------------------------------------------------
class DuplicateRecordID(Exception):
    
    def __init__(self, identifier):
        self.identifier = identifier
        
    def __str__(self):
        return f"In the Dictionary JSON there is a duplicate Record identifier: {self.identifier}"
    
# --------------------------------------------------------------------------------------------------------------------
# class DataProductWriter
#
# Description:
#   This is the main class that processes the data.  It is a container for managing global variables
#
# --------------------------------------------------------------------------------------------------------------------
class DataProductWriter:
    def __init__(self, jsonDict, binaryFileName):
        self.jsonDict = jsonDict
        self.binaryFileName = binaryFileName
        self.totalBytesRead = 0
        self.calculatedCRC = 0
        self.headerJSON = None


    # ----------------------------------------------------------------------------------------------
    # Function: read_and_deserialize
    #
    # Description: 
    #   Reads specified bytes from a binary file, updates CRC, increments byte count,
    #   and deserializes bytes into an integer.
    #
    # Parameters:
    #   nbytes (int): Number of bytes to read.
    #   intType (IntegerType): Integer type for deserialization.
    #
    # Returns:
    #   int: Deserialized integer.
    #
    # Exceptions:
    #   IOError: If reading specified bytes fails.
    #   KeyError: If intType is unrecognized in type_mapping.
    # ----------------------------------------------------------------------------------------------

    def read_and_deserialize(self, nbytes: int, intType: IntegerType) -> int:

        bytes_read = self.binaryFile.read(nbytes)
        if len(bytes_read) != nbytes:
            raise IOError(f"Tried to read {nbytes} bytes from the binary file, but failed.")

        self.calculatedCRC = crc32(bytes_read, self.calculatedCRC) & 0xffffffff
        self.totalBytesRead += nbytes

        try:
            format_str = f'{BIG_ENDIAN}{type_mapping[intType.name]}'
        except KeyError:
            raise KeyError(f"Unrecognized JSON Dictionary Type: {intType}")
        data = struct.unpack(format_str, bytes_read)[0]
        return data

    def read_and_deserialize_string(self) -> str:
        size_store_type_bytes = self.headerJSON.get_size_store_bytes()
        bytes_read_store = self.binaryFile.read(size_store_type_bytes)
        if len(bytes_read_store) != size_store_type_bytes:
            raise IOError(f"Tried to read {size_store_type_bytes} bytes from the binary file, but failed.")

        self.totalBytesRead += size_store_type_bytes
        
        format_str = f'{BIG_ENDIAN}H'
        string_size_data = struct.unpack(format_str, bytes_read_store)[0]

        bytes_read = self.binaryFile.read(string_size_data)
        if len(bytes_read) != string_size_data:
            raise IOError(f"Tried to read {string_size_data} bytes from the binary file, but failed.")

        self.calculatedCRC = crc32(bytes_read_store + bytes_read, self.calculatedCRC) & 0xffffffff
        self.totalBytesRead += string_size_data

        format_str = f'{BIG_ENDIAN}{string_size_data}s'
        data = struct.unpack(format_str, bytes_read)[0]
        data = data.decode()
        return data

    # -----------------------------------------------------------------------------------------------------------------------
    # Function: get_struct_type
    #
    # Description: 
    #   Searches for a structure with a matching identifier in a list of type definitions
    #   and returns the matching structure if found.
    #
    # Parameters:
    #   typeList (List[TypeDef]): A list of type definitions to search through.
    #   identifier (str): The identifier to match against the qualifiedName attribute of each structure.
    #
    # Returns:
    #   TypeDef: The structure that matches the identifier, or None if no match is found.
    #
    # Exceptions:
    #   No explicit exceptions are raised by this function.
    # -----------------------------------------------------------------------------------------------------------------------

    def get_struct_type(self, typeList: List[TypeDef], identifier: str) -> TypeDef:

        for structure in typeList:
            if structure.qualifiedName == identifier:
                return structure

        return None

    # -----------------------------------------------------------------------------------------------------------------------
    # Function: read_field
    #
    # Description: 
    #   Reads and deserializes a field from a binary file, determining the field's size and type
    #   based on the provided configuration, which may be an integer, float, or boolean.
    #
    # Parameters:
    #   field_config (Union[IntegerType, FloatType, BoolType]): Configuration specifying the type and size
    #   of the field to read.
    #
    # Returns:
    #   Union[int, float, bool]: The deserialized value of the field, which can be an integer, float, or boolean.
    #
    # Exceptions:
    #   AssertionError: If the field_config is not an IntegerType, FloatType, or BoolType.
    # -----------------------------------------------------------------------------------------------------------------------

    def read_field(self, field_config: Union[IntegerType, FloatType, BoolType, StringType]) -> Union[int, float, bool]:

        if type(field_config) is IntegerType:
            sizeBytes = field_config.size // 8

        elif type(field_config) is FloatType:
            sizeBytes = field_config.size // 8

        elif type(field_config) is BoolType:
            sizeBytes = field_config.size // 8

        elif type(field_config) is StringType:
            return self.read_and_deserialize_string()

        else:
            assert False, "Unsupported typeKind encountered"

        return self.read_and_deserialize(sizeBytes, field_config)


    # -----------------------------------------------------------------------------------------------------------------------
    # Function: get_struct_item
    #
    # Description: 
    #   This function recursively reads and processes a field from a binary file, adding it to a parent dictionary.
    #   The process varies depending on the field's type:
    #   - For basic types (IntegerType, FloatType, BoolType), it directly reads and assigns the value.
    #   - For EnumType, it reads the value, finds the corresponding enum identifier, and assigns it.
    #   - For AliasType, it reads and assigns the value based on the alias' underlying type.
    #   - For ArrayType, it creates a list, iteratively fills it with elements read recursively, and assigns the list.
    #   - For StructType, it constructs a nested dictionary by recursively processing each struct member.
    #   - For QualifiedType, it resolves the actual type from typeList and recursively processes the field.
    #   This approach allows the function to handle complex, nested data structures by adapting to the field's type,
    #   ensuring each is read and stored appropriately in the parent dictionary.
    #
    # Parameters:
    #   field_name (str): The name of the field to be read and added to the dictionary.
    #   typeKind (TypeKind): The type information of the field, determining how it should be read.
    #   typeList (List[TypeDef]): A list of type definitions, used for resolving qualified types.
    #   parent_dict (Dict[str, int]): The dictionary to which the read field value will be added.
    #
    # Returns:
    #   None: The function does not return a value but modifies parent_dict in place.
    #
    # Exceptions:
    #   AssertionError: If an unsupported typeKind is encountered.

    #
    # -----------------------------------------------------------------------------------------------------------------------

    def get_struct_item(self, field_name: str, typeKind: TypeKind, typeList: List[TypeDef], parent_dict: Dict[str, int]):

        if isinstance(typeKind, IntegerType):
            parent_dict[field_name] = self.read_field(typeKind)

        elif isinstance(typeKind, FloatType):
            parent_dict[field_name] = self.read_field(typeKind)

        elif isinstance(typeKind, BoolType):
            parent_dict[field_name] = self.read_field(typeKind)

        elif isinstance(typeKind, StringType):
            parent_dict[field_name] = self.read_field(typeKind)

        elif isinstance(typeKind, EnumType):
            value = self.read_field(typeKind.representationType)
            enum_mapping = typeKind.enumeratedConstants
            reverse_mapping = {enum.value: enum.name for enum in enum_mapping}
            parent_dict[field_name] = reverse_mapping[value]

        elif isinstance(typeKind, AliasType):
            self.get_struct_item(field_name, typeKind.underlyingType, typeList, parent_dict)

        elif isinstance(typeKind, ArrayType):
            array_list = []
            for item in range(typeKind.size):
                element_dict = {} 
                self.get_struct_item("arrayElement", typeKind.elementType, typeList, element_dict)
                array_list.append(element_dict["arrayElement"]) 
            parent_dict[field_name] = array_list

        elif isinstance(typeKind, StructType):
            array_list = []
            sorted_members = dict(sorted(typeKind.members.items(), key=lambda member: member[1].index))
            for key, member in sorted_members.items():
                for i in range(member.size):
                    element_dict = {}
                    self.get_struct_item(key, member.type, typeList, element_dict)
                    #array_list.append(element_dict[key])
                    array_list.append(element_dict)
                parent_dict[field_name] = array_list

        elif isinstance(typeKind, QualifiedType):
            qualType = self.get_struct_type(typeList, typeKind.name)
            self.get_struct_item(field_name, qualType, typeList, parent_dict)

        else:
            assert False, "Unsupported typeKind encountered"


    # -----------------------------------------------------------------------------------------------------------------------
    # Function: get_header_info
    #
    # Description: 
    #   Extracts header information from a given DPHeader object, populating and returning a dictionary with the data.
    #   Iterates over header fields, reading each and updating a root dictionary with the field values. After processing all fields,
    #   it reads and compares the header hash with a computed CRC value to verify data integrity. If the CRC check fails, it raises
    #   a CRCError. This function demonstrates a pattern of custom exceptions to manage and validate
    #   binary data parsing.
    #
    # Parameters:
    #   headerJSON (DPHeader): The DPHeader object containing header information and type definitions.
    #
    # Returns:
    #   Dict[str, int]: A dictionary populated with the header fields and their corresponding values.
    #
    # Exceptions:
    #   CRCError: Raised if the computed CRC does not match the expected header hash value.
    # -----------------------------------------------------------------------------------------------------------------------

    def get_header_info(self, headerJSON: DPHeader) -> Dict[str, int]:

        header_fields = headerJSON.header
        rootDict = {}

        for field_name, field_info in header_fields.items():
            # Header is composed of enums, arrays, and aliases
            if isinstance(field_info, EnumType):
                self.get_struct_item(field_name, field_info.representationType, headerJSON.typeDefinitions, rootDict)
            elif isinstance(field_info, ArrayType):
                self.get_struct_item(field_name, field_info, headerJSON.typeDefinitions, rootDict)
            else:
                self.get_struct_item(field_name, field_info.type, headerJSON.typeDefinitions, rootDict)

        computedHash = self.calculatedCRC
        rootDict['headerHash'] = self.read_field(headerJSON.headerHash.type)
        self.calculatedCRC = 0

        if rootDict['headerHash'] != computedHash:
            raise CRCError("Header", rootDict['headerHash'], computedHash)

        return rootDict

    # ------------------------------------------------------------------------------------------
    # Function: get_record_data
    #
    # Description:
    #     Retrieves and processes the record data based on a given header and dictionary
    #     definition. The function first reads the 'dataId' from the header to identify the
    #     relevant record. It then processes the record's data, handling both scalar values and
    #     arrays by reading each item according to its type. For arrays, it also reads the
    #     'dataSize' from the header to determine the number of items to process.
    #
    # Parameters:
    #     - headerJSON (DPHeader): An object containing the header information, including
    #       identifiers and sizes for the data to be processed.
    #     - dictJSON (FprimeDict): An object containing definitions for records and types,
    #       which are used to process the data correctly.
    #
    # Returns:
    #     Dict[str, int]: A dictionary with the processed data, including 'dataId', optionally
    #     'size' for arrays, and the data itself. For arrays, the data is indexed by its position
    #     within the array.

    # ------------------------------------------------------------------------------------------

    def get_record_data(self, headerJSON: DPHeader, dictJSON: FprimeDict) -> Dict[str, int]:
        rootDict = {}
        # Go through all the Records and find the one that matches recordId
        rootDict['dataId'] = self.read_field(headerJSON.dataId.underlyingType)
        for record in dictJSON.records:
            if record.id == rootDict['dataId']:    
                print(f'Processing Record ID {record.id}')
                if record.array:
                    dataSize = self.read_field(headerJSON.dataSize.type)
                    rootDict['size'] = dataSize
                    array_data = []
                    for i in range(dataSize):
                        element_dict = {}
                        self.get_struct_item("arrayElement", record.type, dictJSON.typeDefinitions, element_dict)
                        array_data.append(element_dict["arrayElement"])
                    rootDict['data'] = array_data
                else:
                    # For non-array records, directly use 'data' as the key.
                    self.get_struct_item("data", record.type, dictJSON.typeDefinitions, rootDict)

                return rootDict
        raise RecordIDNotFound(rootDict['dataId'])

    

    # --------------------------------------------------------------------------------------------------------------------
    # Function handleException
    # 
    # Description: 
    #   Function for handling exceptions by displaying an error message and terminating the program.
    #   It displays the provided exception message with color-coded output for emphasis. 
    #   After displaying the messages, it immediately exits the program, halting further execution.
    #
    # Parameters:
    #   msg (str): The specific error message to be displayed.
    #
    # Returns:
    #   None: This function does not return a value. It terminates the program execution with sys.exit().
    #
    # Exceptions:
    #   No explicit exceptions are raised by this function, but it triggers the program's termination.
    # -----------------------------------------------------------------------------------------------------------------
    def handleException(self, msg):
        errorMessage = f"*** Error in processing: "
        print(bcolors.FAIL)
        print(errorMessage)
        print(bcolors.WARNING)
        print(msg)
        print(bcolors.ENDC)
        sys.exit()


    # -------------------------------------------------------------------------------------------------------------------------
    # Function check_record_data
    #
    # Description: 
    #   Validates record data in a given dictionary JSON object by ensuring there are no duplicate record identifiers.
    #   Iterates through the records in the dictionary, checking each record's identifier against a set that 
    #   tracks unique identifiers.  If a duplicate identifier is detected, a DuplicateRecordID exception is raised.
    #
    # Parameters:
    #   dictJSON (FprimeDict): The dictionary JSON object containing records to be validated.
    #
    # Returns:
    #   None: This function does not return a value. It either completes successfully or raises an exception.
    #
    # Exceptions:
    #   DuplicateRecordID: Raised if a duplicate record identifier is found in the dictionary JSON object.

    # -----------------------------------------------------------------------------------------------------------------------
    def check_record_data(self, dictJSON: FprimeDict):
        idSet = set()
        for record in dictJSON.records:
            if record.id in idSet:
                raise DuplicateRecordID(record.id)
            else:
                idSet.add(record.id)



    # -------------------------------------------------------------------------------------------------------------------------
    # Function process
    #
    # Description: 
    #   Main processing
    # -------------------------------------------------------------------------------------------------------------------------
    def process(self):

        try:

            # Read the F prime JSON dictionary
            print(f"Parsing {self.jsonDict}...")
            try:
                with open(self.jsonDict, 'r') as fprimeDictFile:
                    dict_json = json.load(fprimeDictFile)
                    dictJSON = FprimeDict(**dict_json)

                    header_json = header_hash_data
                    if "typeDefinitions" in dict_json:
                        header_json["typeDefinitions"] = dict_json["typeDefinitions"]
                    if "constants" in dict_json:
                        header_json["constants"] = dict_json["constants"]
                    self.headerJSON = DPHeader(**header_json)
            
            except json.JSONDecodeError as e:
                raise DictionaryError(self.jsonDict, e.lineno)
            
            self.check_record_data(dictJSON)

            with open(self.binaryFileName, 'rb') as self.binaryFile:

                # Read the header data up until the Records
                headerData = self.get_header_info(self.headerJSON)

                # Read the total data size
                dataSize = headerData['DataSize']

                # Restart the count of bytes read
                self.totalBytesRead = 0

                recordList = [headerData]

                while self.totalBytesRead < dataSize:
                    recordData = self.get_record_data(self.headerJSON, dictJSON)
                    recordList.append(recordData)

                computedCRC = self.calculatedCRC
                # Read the data checksum
                headerData['dataHash'] = self.read_field(self.headerJSON.dataHash.type)

                if computedCRC != headerData['dataHash']:
                    raise CRCError("Data", headerData['dataHash'], computedCRC)


        except (FileNotFoundError, RecordIDNotFound, IOError, KeyError, json.JSONDecodeError, 
                DictionaryError, CRCError, DuplicateRecordID) as e:
            self.handleException(e)

        except (ValueError) as e:
            error = e.errors()[0]
            msg = f'ValueError in JSON file {error["loc"]}: {error["msg"]}'
            self.handleException(msg)


        # Output the generated json to a file
        baseName = os.path.basename(self.binaryFileName)
        outputJsonFile = os.path.splitext(baseName)[0] + '.json'
        if outputJsonFile.startswith('._'):
            outputJsonFile = outputJsonFile.replace('._', '')
        with open(outputJsonFile, 'w') as file:
            json.dump(recordList, file, indent=2)

        print(f'Output data generated in {outputJsonFile}')


# ------------------------------------------------------------------------------------------
# main program
#
# ------------------------------------------------------------------------------------------
def main():
    args = parse_args()
    DataProductWriter(args.jsonDict, args.binFile).process()

if __name__ == "main":
    sys.exit(main())





"""
@brief Base data template class.

Data templates are classes whose instances describe a specific class of data
items. For example, a data template instance could describe the AF_ASSERT_0
event or the channel FR_CycleTime.

@date Created July 2, 2018
@author R. Joseph Paetz

@bug No known bugs
"""
class DataTemplate:
    """Base class for all data template classes"""

    def __init__(self):
        """
        Constructor.

        The only required fields for template classes are id and name

        Returns:
            Initialized Template object
        """
        # Initialize at least id and name here

    def get_id(self):
        raise NotImplementedError

    def get_name(self):
        raise NotImplementedError
